import unittest2
from predownload import PreDownload


class PreDownloadTest(unittest2.TestCase):
    def test_can_instantiate_predownload(self):
        pdl = PreDownload('something', 'https://example.com')
        self.assertEqual('something', pdl.read())

    def test_raises_when_no_url_is_provided(self):
        self.assertRaises(ValueError, lambda: PreDownload('something', None))



import ujson as json
import os
import threading
import numpy as np
from hammingdb import HammingDb
from zounds.persistence import TimeSliceEncoder, TimeSliceDecoder
from zounds.timeseries import ConstantRateTimeSeries
from zounds.timeseries import TimeSlice


class SearchResults(object):
    def __init__(self, query, results):
        super(SearchResults, self).__init__()
        self.query = query
        self.results = results

    def __iter__(self):
        for result in self.results:
            yield result


class HammingIndex(object):
    def __init__(
            self,
            document,
            feature,
            version=None,
            path='',
            db_size_bytes=1000000000,
            listen=False,
            writeonly=False,
            **extra_data):

        super(HammingIndex, self).__init__()
        self.document = document
        self.feature = feature
        self.db_size_bytes = db_size_bytes
        self.path = path
        self.extra_data = extra_data
        self.writeonly = writeonly

        version = version or self.feature.version

        self.hamming_db_path = os.path.join(
            self.path, 'index.{self.feature.key}.{version}'
                .format(**locals()))

        try:
            self.event_log = document.event_log
        except AttributeError:
            self.event_log = None

        try:
            self.hamming_db = HammingDb(
                self.hamming_db_path, code_size=None, writeonly=self.writeonly)
        except ValueError:
            self.hamming_db = None

        self.encoder = TimeSliceEncoder()
        self.decoder = TimeSliceDecoder()
        self.thread = None

        if listen:
            self.listen()

    def close(self):
        try:
            self.stop()
        except:
            pass

        try:
            self.hamming_db.close()
        except:
            pass

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __len__(self):
        if self.hamming_db is None:
            return 0
        return len(self.hamming_db)

    def stop(self):
        self.event_log.unsubscribe()

    def listen(self):
        self.thread = threading.Thread(target=self._listen)
        self.thread.daemon = True
        self.thread.start()

    def _init_hamming_db(self, code=None):
        if self.hamming_db is not None:
            return
        code_size = len(code) if code else None
        self.hamming_db = HammingDb(
            self.hamming_db_path, code_size=code_size, writeonly=self.writeonly)

    def _synchronously_process_events(self):
        self._listen(raise_when_empty=True)

    def add_all(self):
        for doc in self.document:
            self.add(doc._id)

    def _collect_extra_data(self, _id, ts):
        if not self.extra_data:
            return None

        doc = self.document(_id)
        return dict(
            ((key, func(doc, ts)) for key, func in self.extra_data.iteritems()))

    def add(self, _id, timestamp=''):
        # load the feature from the feature database
        feature = self.feature(_id=_id, persistence=self.document)

        try:
            arr = ConstantRateTimeSeries(feature)
        except ValueError:
            arr = feature

        # extract codes and timeslices from the feature
        for ts, data in arr.iter_slices():
            code = self.encode_query(data)
            encoded_ts = dict(
                _id=_id,
                **self.encoder.dict(ts))
            extra_data = self._collect_extra_data(_id, ts)
            if extra_data:
                encoded_ts['extra_data'] = extra_data
            self._init_hamming_db(code)
            self.hamming_db.append(code, json.dumps(encoded_ts))
            self.hamming_db.set_metadata('timestamp', bytes(timestamp))

    def _listen(self, raise_when_empty=False):

        if self.hamming_db is not None:
            last_timestamp = self.hamming_db.get_metadata('timestamp') or ''
        else:
            last_timestamp = ''

        if not self.event_log:
            raise ValueError(
                '{self.document} must have an event log configured'
                    .format(**locals()))

        subscription = self.event_log.subscribe(
            last_id=last_timestamp, raise_when_empty=raise_when_empty)

        for timestamp, data in subscription:

            # parse the data from the event stream
            data = json.loads(data)
            _id, name, version = data['_id'], data['name'], data['version']

            # ensure that it's about the feature we're subscribed to
            if name != self.feature.key or version != self.feature.version:
                continue

            self.add(_id, timestamp)

    def _parse_result(self, result):
        d = json.loads(result)
        ts = TimeSlice(**self.decoder.kwargs(d))

        if not self.extra_data:
            return d['_id'], ts

        return d['_id'], ts, d['extra_data']

    def decode_query(self, binary_query):
        packed = np.fromstring(binary_query, dtype=np.uint8)
        return np.unpackbits(packed)

    def encode_query(self, feature):
        if isinstance(feature, str):
            return feature
        elif feature.dtype == np.uint64:
            return feature.tostring()
        elif feature.dtype == np.uint8 or feature.dtype == np.bool:
            return np.packbits(feature).tostring()
        else:
            raise ValueError(
                'feature must be a raw bit string, an already packed uint64'
                'array, or an "unpacked" uint8 or bool array')

    def random_search(self, n_results, multithreaded=False, sort=False):
        self._init_hamming_db()
        code, raw_results = self.hamming_db.random_search(
            n_results, multithreaded, sort=sort)
        parsed_results = (self._parse_result(r) for r in raw_results)
        return SearchResults(code, parsed_results)

    def search(self, feature, n_results, multithreaded=False, sort=False):
        self._init_hamming_db()
        code = self.encode_query(feature)
        raw_results = self.hamming_db.search(
            code, n_results, multithreaded, sort=sort)
        parsed_results = (self._parse_result(r) for r in raw_results)
        return SearchResults(code, parsed_results)



from preprocess import Preprocessor, PreprocessResult


class SklearnModel(Preprocessor):
    def __init__(self, model=None, needs=None):
        super(SklearnModel, self).__init__(needs=needs)
        self.model = model

    def _forward_func(self):
        def x(d, model=None):
            from zounds.core import ArrayWithUnits, IdentityDimension
            transformed = model.transform(d.reshape((d.shape[0], -1)))
            try:
                return ArrayWithUnits(
                    transformed, (d.dimensions[0], IdentityDimension()))
            except AttributeError:
                return transformed

        return x

    def _backward_func(self):
        def x(d, model=None, shape=None):
            return model.inverse_transform(d).reshape((-1,) + shape)

        return x

    def _process(self, data):
        data = self._extract_data(data)
        model = self.model.fit(data.reshape((data.shape[0], -1)))
        shape = data.shape[1:]
        op = self.transform(model=model)
        inv_data = self.inversion_data(model=model, shape=shape)
        inv = self.inverse_transform()
        data = op(data)
        model_cls = self.model.__class__.__name__
        yield PreprocessResult(
            data,
            op,
            inversion_data=inv_data,
            inverse=inv,
            name='SklearnModel.{model_cls}'.format(**locals()))


class WithComponents(SklearnModel):
    def __init__(self, model=None, needs=None):
        super(WithComponents, self).__init__(model=model, needs=needs)

    def _backward_func(self):
        def x(d, model=None, shape=None):
            import numpy as np
            return np.dot(d, model.components_).reshape((-1,) + shape)
        return x



from io import BytesIO

from api import ZoundsApp
import tornado.websocket
import tornado.web
import json
from collections import defaultdict

import matplotlib

matplotlib.use('Agg')
from matplotlib import pyplot as plt


class TrainingMonitorApp(ZoundsApp):
    def __init__(
            self,
            trainer,
            keys_to_graph,
            batch_frequency=10,
            n_training_points=100,
            epoch_key='epoch',
            batch_key='batch',
            base_path=r'/zounds/',
            model=None,
            visualization_feature=None,
            audio_feature=None,
            globals={},
            locals={},
            secret=None):

        super(TrainingMonitorApp, self).__init__(
            base_path=base_path,
            model=model,
            visualization_feature=visualization_feature,
            audio_feature=audio_feature,
            globals=globals,
            locals=locals,
            html='training_monitor.html',
            secret=secret)

        self.n_training_points = n_training_points
        self.batch_frequency = batch_frequency
        self.batch_key = batch_key
        self.epoch_key = epoch_key
        self.keys_to_graph = keys_to_graph
        self.trainer = trainer
        self.training_history = defaultdict(list)
        self.trainer.register_batch_complete_callback(
            self._collect_training_history)

    def custom_routes(self):
        routes = super(TrainingMonitorApp, self).custom_routes()
        routes.extend([
            (r'/zounds/training/?', self.training_handler()),
            (r'/zounds/graph/?', self.graph_handler())
        ])
        return routes

    def _collect_training_history(self, *args, **kwargs):
        batch = kwargs['batch']

        if batch % self.batch_frequency:
            return

        for k in self.keys_to_graph:
            # truncate
            self.training_history[k] = \
                self.training_history[k][-self.n_training_points:]
            # append the new data
            self.training_history[k].append(kwargs[k])

    def training_handler(self):
        app = self

        class TrainingHandler(tornado.websocket.WebSocketHandler):
            def _send_message(self):
                def x(*args, **kwargs):
                    batch = kwargs['batch']

                    if batch % app.batch_frequency:
                        return

                    data = dict(epoch=kwargs['epoch'], batch=batch)
                    for key in app.keys_to_graph:
                        data[key] = kwargs[key]
                    self.write_message(json.dumps(data))

                return x

            def open(self):
                self.func = self._send_message()
                app.trainer \
                    .register_batch_complete_callback(self.func)

            def on_close(self):
                app.trainer \
                    .unregister_batch_complete_callback(self.func)

        return TrainingHandler

    def graph_handler(self):
        app = self

        class GraphHandler(tornado.web.RequestHandler):
            def get(self):
                plt.style.use('dark_background')

                fig = plt.figure()
                handles = []
                for k in app.keys_to_graph:
                    handle, = plt.plot(app.training_history[k], label=k)
                    handles.append(handle)
                plt.legend(handles=handles)

                bio = BytesIO()
                plt.savefig(
                    bio, bbox_inches='tight', pad_inches=0, format='png')
                bio.seek(0)
                fig.clf()
                plt.close('all')
                self.set_header('Content-Type', 'image/png')
                self.write(bio.read())
                self.finish()

        return GraphHandler


class SupervisedTrainingMonitorApp(TrainingMonitorApp):
    def __init__(
            self,
            trainer,
            batch_frequency=10,
            n_training_points=100,
            epoch_key='epoch',
            batch_key='batch',
            base_path=r'/zounds',
            model=None,
            visualization_feature=None,
            audio_feature=None,
            globals={},
            locals={},
            secret=None):
        super(SupervisedTrainingMonitorApp, self).__init__(
            trainer=trainer,
            keys_to_graph=('train_error', 'test_error'),
            model=model,
            batch_frequency=batch_frequency,
            n_training_points=n_training_points,
            epoch_key=epoch_key,
            batch_key=batch_key,
            base_path=base_path,
            visualization_feature=visualization_feature,
            audio_feature=audio_feature,
            globals=globals,
            locals=locals,
            secret=secret)


class GanTrainingMonitorApp(TrainingMonitorApp):
    def __init__(
            self,
            trainer,
            batch_frequency=10,
            n_training_points=100,
            epoch_key='epoch',
            batch_key='batch',
            base_path=r'/zounds',
            model=None,
            visualization_feature=None,
            audio_feature=None,
            globals={},
            locals={},
            secret=None):
        super(GanTrainingMonitorApp, self).__init__(
            trainer=trainer,
            keys_to_graph=('generator_score', 'real_score', 'critic_loss'),
            model=model,
            batch_frequency=batch_frequency,
            n_training_points=n_training_points,
            epoch_key=epoch_key,
            batch_key=batch_key,
            base_path=base_path,
            visualization_feature=visualization_feature,
            audio_feature=audio_feature,
            globals=globals,
            locals=locals,
            secret=secret)


class TripletEmbeddingMonitorApp(TrainingMonitorApp):
    def __init__(
            self,
            trainer,
            batch_frequency=10,
            n_training_points=100,
            epoch_key='epoch',
            batch_key='batch',
            base_path=r'/zounds',
            model=None,
            visualization_feature=None,
            audio_feature=None,
            globals={},
            locals={},
            secret=None):
        super(TripletEmbeddingMonitorApp, self).__init__(
            trainer=trainer,
            keys_to_graph=('error',),
            model=model,
            batch_frequency=batch_frequency,
            n_training_points=n_training_points,
            epoch_key=epoch_key,
            batch_key=batch_key,
            base_path=base_path,
            visualization_feature=visualization_feature,
            audio_feature=audio_feature,
            globals=globals,
            locals=locals,
            secret=secret)



from resample import Resample
from zounds.timeseries import SR44100, SR11025, Seconds
from zounds.synthesize import SilenceSynthesizer
from multiprocessing.pool import ThreadPool
import unittest2


class ResampleTests(unittest2.TestCase):

    def test_can_do_multithreaded_resampling(self):
        synth = SilenceSynthesizer(SR44100())
        audio = [synth.synthesize(Seconds(5)) for _ in xrange(10)]
        pool = ThreadPool(4)

        def x(samples):
            rs = Resample(int(SR44100()), int(SR11025()))
            return rs(samples, end_of_input=True)

        resampled = pool.map(x, audio)
        self.assertEqual(10, len(resampled))

    def test_correct_output_with_stereo(self):
        synth = SilenceSynthesizer(SR44100())
        samples = synth.synthesize(Seconds(1)).stereo
        rs = Resample(int(samples.samplerate), int(SR11025()), nchannels=2)
        resampled = rs(samples, end_of_input=True)
        self.assertEqual((11025, 2), resampled.shape)


from __future__ import division
cimport numpy as np
import numpy as np

INT_DTYPE = np.int
ctypedef np.int_t INT_DTYPE_t

FLOAT_DTYPE = np.float32
ctypedef np.float32_t FLOAT_DTYPE_t

UINT64_DTYPE = np.uint64
ctypedef np.uint64_t UINT64_DTYPE_t

ctypedef unsigned long ULong

cimport cython

cdef extern int __builtin_popcount(unsigned int) nogil
cdef extern int __builtin_popcountl(unsigned long) nogil
cdef extern int __builtin_popcountll(unsigned long long) nogil

@cython.boundscheck(False)
@cython.wraparound(False)
def count_bits(np.ndarray[UINT64_DTYPE_t,ndim = 1] n):
    cdef np.ndarray[INT_DTYPE_t,ndim = 1] out = \
        np.ndarray(n.size,dtype = INT_DTYPE)
    cdef int i = 0
    cdef int l = n.size
    cdef UINT64_DTYPE_t a = 0x5555555555555555
    cdef UINT64_DTYPE_t b = 0x3333333333333333
    cdef UINT64_DTYPE_t c = 0xF0F0F0F0F0F0F0F
    cdef UINT64_DTYPE_t d = 0x101010101010101
    cdef UINT64_DTYPE_t z = 0
    for i in range(l):
        z = n[i]
        z = z - ((z >> 1) & a)
        z = (z & b) + ((z >> 2) & b)
        out[i] = ((z + (z >> 4)) & c) * d >> 56
    return out


@cython.boundscheck(False)
@cython.wraparound(False)
def count_packed_bits(np.ndarray[UINT64_DTYPE_t,ndim = 2] n):
    cdef int ns = n.shape[0]
    cdef int ns2 = n.shape[1]
    cdef np.ndarray[INT_DTYPE_t,ndim = 1] out = \
        np.ndarray(ns,dtype = INT_DTYPE)
    cdef int i = 0
    cdef int j = 0
    cdef UINT64_DTYPE_t q = 0
    cdef UINT64_DTYPE_t a = 0x5555555555555555
    cdef UINT64_DTYPE_t b = 0x3333333333333333
    cdef UINT64_DTYPE_t c = 0xF0F0F0F0F0F0F0F
    cdef UINT64_DTYPE_t d = 0x101010101010101
    cdef UINT64_DTYPE_t z = 0
    for i in range(ns):
        z = 0
        for j in range(ns2):
            q = n[i, j]
            q = q - ((q >> 1) & a)
            q = (q & b) + ((q >> 2) & b)
            z += ((q + (q >> 4)) & c) * d >> 56
        out[i] = z
    return out


from __future__ import division
from frequencyscale import LinearScale, FrequencyBand, ExplicitScale
from tfrepresentation import FrequencyDimension
from frequencyadaptive import FrequencyAdaptive
from zounds.timeseries import \
    audio_sample_rate, TimeSlice, Seconds, TimeDimension, HalfLapped
from zounds.core import ArrayWithUnits, IdentityDimension
from sliding_window import IdentityWindowingFunc, HanningWindowingFunc
from zounds.loudness import log_modulus, unit_scale
import numpy as np
from scipy.signal import resample, firwin2
from matplotlib import cm


def fft(x, axis=-1, padding_samples=0):
    """
    Apply an FFT along the given dimension, and with the specified amount of
    zero-padding

    Args:
        x (ArrayWithUnits): an :class:`~zounds.core.ArrayWithUnits` instance
            which has one or more :class:`~zounds.timeseries.TimeDimension`
            axes
        axis (int): The axis along which the fft should be applied
        padding_samples (int): The number of padding zeros to apply along
            axis before performing the FFT
    """
    if padding_samples > 0:
        padded = np.concatenate(
            [x, np.zeros((len(x), padding_samples), dtype=x.dtype)],
            axis=axis)
    else:
        padded = x

    transformed = np.fft.rfft(padded, axis=axis, norm='ortho')
    sr = audio_sample_rate(int(Seconds(1) / x.dimensions[axis].frequency))
    scale = LinearScale.from_sample_rate(sr, transformed.shape[-1])
    new_dimensions = list(x.dimensions)
    new_dimensions[axis] = FrequencyDimension(scale)
    return ArrayWithUnits(transformed, new_dimensions)


def stft(x, window_sample_rate=HalfLapped(), window=HanningWindowingFunc()):
    duration = TimeSlice(window_sample_rate.duration)
    frequency = TimeSlice(window_sample_rate.frequency)
    _, arr = x.sliding_window_with_leftovers(
        duration, frequency, dopad=True)
    window = window or IdentityWindowingFunc()
    windowed = arr * window
    return fft(windowed)


def phase_shift(coeffs, samplerate, time_shift, axis=-1, frequency_band=None):
    frequency_dim = coeffs.dimensions[axis]
    if not isinstance(frequency_dim, FrequencyDimension):
        raise ValueError(
            'dimension {axis} of coeffs must be a FrequencyDimension instance, '
            'but was {cls}'.format(axis=axis, cls=frequency_dim.__class__))

    n_coeffs = coeffs.shape[axis]
    shift_samples = int(time_shift / samplerate.frequency)
    shift = (np.arange(0, n_coeffs) * 2j * np.pi) / n_coeffs
    shift = np.exp(-shift * shift_samples)
    shift = ArrayWithUnits(shift, [frequency_dim])

    frequency_band = frequency_band or slice(None)
    new_coeffs = coeffs.copy()

    if coeffs.ndim == 1:
        new_coeffs[frequency_band] *= shift[frequency_band]
        return new_coeffs

    slices = [slice(None) for _ in xrange(coeffs.ndim)]
    slices[axis] = frequency_band
    new_coeffs[tuple(slices)] *= shift[frequency_band]
    return new_coeffs


def apply_scale(short_time_fft, scale, window=None):
    magnitudes = np.abs(short_time_fft.real)
    spectrogram = scale.apply(magnitudes, window)
    dimensions = short_time_fft.dimensions[:-1] + (FrequencyDimension(scale),)
    return ArrayWithUnits(spectrogram, dimensions)


def rainbowgram(time_frequency_repr, colormap=cm.rainbow):
    # magnitudes on a log scale, and shifted and
    # scaled to the unit interval
    magnitudes = np.abs(time_frequency_repr.real)
    magnitudes = log_modulus(magnitudes * 1000)
    magnitudes = unit_scale(magnitudes)

    angles = np.angle(time_frequency_repr)
    angles = np.unwrap(angles, axis=0)
    angles = np.gradient(angles)[0]
    angles = unit_scale(angles)

    colors = colormap(angles)
    colors *= magnitudes[..., None]

    # exclude the alpha channel, if there is one
    colors = colors[..., :3]
    arr = ArrayWithUnits(
        colors, time_frequency_repr.dimensions + (IdentityDimension(),))
    return arr


def fir_filter_bank(scale, taps, samplerate, window):
    basis = np.zeros((len(scale), taps))
    basis = ArrayWithUnits(basis, [
        FrequencyDimension(scale),
        TimeDimension(*samplerate)])

    nyq = samplerate.nyquist

    for i, band in enumerate(scale):
        freqs = np.linspace(
            band.start_hz / nyq, band.stop_hz / nyq, len(window))
        freqs = [0] + list(freqs) + [1]
        gains = [0] + list(window) + [0]
        basis[i] = firwin2(taps, freqs, gains)

    return basis


def dct_basis(size):
    r = np.arange(size)
    basis = np.outer(r, r + 0.5)
    basis = np.cos((np.pi / size) * basis)
    return basis


def frequency_decomposition(x, sizes):
    sizes = sorted(sizes)

    if x.ndim == 1:
        end = x.dimensions[0].end
        x = ArrayWithUnits(
            x[None, ...], [TimeDimension(end, end), x.dimensions[0]])

    original_size = x.shape[-1]
    time_dimension = x.dimensions[-1]
    samplerate = audio_sample_rate(time_dimension.samples_per_second)
    data = x.copy()

    bands = []
    frequency_bands = []
    start_hz = 0

    for size in sizes:
        if size != original_size:
            s = resample(data, size, axis=-1)
        else:
            s = data

        bands.append(s)
        data -= resample(s, original_size, axis=-1)

        stop_hz = samplerate.nyquist * (size / original_size)
        frequency_bands.append(FrequencyBand(start_hz, stop_hz))
        start_hz = stop_hz

    scale = ExplicitScale(frequency_bands)
    return FrequencyAdaptive(bands, scale=scale, time_dimension=x.dimensions[0])



from __future__ import print_function, division
import numpy as np
import featureflow as ff
import hashlib


def simple_settings(cls):
    """
    Create sane default persistence settings for learning pipelines
    :param cls: The class to decorate
    """

    class Settings(ff.PersistenceSettings):
        _id = cls.__name__
        id_provider = ff.StaticIdProvider(_id)
        key_builder = ff.StringDelimitedKeyBuilder()
        database = ff.FileSystemDatabase(
            path=_id, key_builder=key_builder, createdirs=True)

    class Model(cls, Settings):
        pass

    Model.__name__ = cls.__name__
    Model.__module__ = cls.__module__
    return Model


def object_store_pipeline_settings(container, region, username, api_key):
    def decorator(cls):
        class Settings(ff.PersistenceSettings):
            _id = cls.__name__
            id_provider = ff.StaticIdProvider(_id)
            key_builder = ff.StringDelimitedKeyBuilder()
            database = ff.ObjectStoreDatabase(
                container, username, api_key, region, key_builder=key_builder)

        class Model(cls, Settings):
            pass

        Model.__name__ = cls.__name__
        Model.__module__ = cls.__module__
        return Model

    return decorator


def model_hash(model):
    h = hashlib.md5()
    h.update(str(model))
    for p in model.parameters():
        h.update(p.data.cpu().numpy())
    return h.hexdigest()


try:
    import torch
    from torch import nn
    from torch.nn import functional as F
    from torch.autograd import Variable


    def to_var(x, volatile=False):
        t = torch.from_numpy(x)
        v = Variable(t, volatile=volatile).cuda()
        return v


    def from_var(x):
        return x.data.cpu().numpy()


    def try_network(network, x, **kwargs):
        network = network.cuda()
        x = to_var(x)
        result = network(x, **kwargs)
        return result


    def apply_network(network, x, chunksize=None):
        """
        Apply a pytorch network, potentially in chunks
        """
        x = to_var(x, volatile=True)

        if chunksize is None:
            return from_var(network(x))

        return np.concatenate(
            [from_var(network(x[i: i + chunksize]))
             for i in xrange(0, len(x), chunksize)])


    def sample_norm(x):
        """
        pixel norm as described in section 4.2 here:
        https://arxiv.org/pdf/1710.10196.pdf
        """
        original = x
        # square
        x = x ** 2
        # feature-map-wise sum
        x = torch.sum(x, dim=1)
        # scale by number of feature maps
        x *= 1.0 / original.shape[1]
        x += 10e-8
        x = torch.sqrt(x)
        return original / x.view(-1, 1, x.shape[-1])


    def feature_map_size(inp, kernel, stride=1, padding=0):
        return ((inp - kernel + (2 * padding)) / stride) + 1


    class ConvLayer(nn.Module):
        def __init__(
                self,
                layer_type,
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                activation=lambda x: F.leaky_relu(x, 0.2),
                dropout=True,
                batch_norm=True,
                dilation=1,
                sample_norm=False):

            super(ConvLayer, self).__init__()
            self.sample_norm = sample_norm
            self.dropout = dropout
            self.l1 = layer_type(
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                bias=False,
                dilation=dilation)

            self.bn = None

            if batch_norm:
                if '1d' in layer_type.__name__:
                    self.bn = nn.BatchNorm1d(out_channels)
                else:
                    self.bn = nn.BatchNorm2d(out_channels)

            self.activation = activation

        @property
        def out_channels(self):
            return self.l1.out_channels

        @property
        def in_channels(self):
            return self.l1.in_channels

        @property
        def kernel_size(self):
            return self.l1.kernel_size

        @property
        def stride(self):
            return self.l1.stride

        @property
        def padding(self):
            return self.l1.padding

        def forward(self, x):
            x = self.l1(x)

            if self.sample_norm:
                x = sample_norm(x)
            elif self.bn:
                x = self.bn(x)

            if self.activation:
                x = self.activation(x)

            if self.dropout:
                x = F.dropout(x, 0.2, self.training)
            return x


    class Conv1d(ConvLayer):
        def __init__(
                self,
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                dropout=True,
                batch_norm=True,
                dilation=1,
                sample_norm=False,
                activation=lambda x: F.leaky_relu(x, 0.2)):
            super(Conv1d, self).__init__(
                nn.Conv1d,
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                activation=activation,
                dropout=dropout,
                batch_norm=batch_norm,
                dilation=dilation,
                sample_norm=sample_norm)


    class ConvTranspose1d(ConvLayer):
        def __init__(
                self,
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                activation=lambda x: F.leaky_relu(x, 0.2),
                dropout=True,
                batch_norm=True,
                dilation=1,
                sample_norm=False):
            super(ConvTranspose1d, self).__init__(
                nn.ConvTranspose1d,
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                activation=activation,
                dropout=dropout,
                batch_norm=batch_norm,
                dilation=dilation,
                sample_norm=sample_norm)


    class Conv2d(ConvLayer):
        def __init__(
                self,
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                dropout=True,
                batch_norm=True,
                dilation=1,
                sample_norm=False,
                activation=lambda x: F.leaky_relu(x, 0.2), ):
            super(Conv2d, self).__init__(
                nn.Conv2d,
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                activation=activation,
                dropout=dropout,
                batch_norm=batch_norm,
                dilation=dilation,
                sample_norm=sample_norm)


    class ConvTranspose2d(ConvLayer):
        def __init__(
                self,
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                activation=lambda x: F.leaky_relu(x, 0.2),
                dropout=True,
                batch_norm=True,
                dilation=1,
                sample_norm=False):
            super(ConvTranspose2d, self).__init__(
                nn.ConvTranspose2d,
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                activation=activation,
                dropout=dropout,
                batch_norm=batch_norm,
                dilation=dilation,
                sample_norm=sample_norm)


    class FrequencyDecompositionAnalyzer(nn.Module):
        def __init__(
                self,
                first_feature_map_size,
                growth_factor,
                bands,
                bottleneck):

            super(FrequencyDecompositionAnalyzer, self).__init__()
            self.bottleneck = bottleneck
            self.first_feature_map_size = first_feature_map_size
            self.growth_factor = growth_factor
            self.bands = bands
            self.layer_stacks = [[] for _ in bands]
            self.total_bands = sum(bands)

            for i, band in enumerate(bands):
                out_channels = self.first_feature_map_size
                l1 = Conv1d(1, out_channels, 8, 4, 0)
                self.layer_stacks[i].append(l1)
                fms = feature_map_size(
                    band, l1.kernel_size[0], l1.stride[0], l1.padding[0])
                while fms > 1:
                    in_channels = out_channels
                    out_channels = int(out_channels * self.growth_factor)
                    l1 = Conv1d(in_channels, out_channels, 3, 2, 0)
                    self.layer_stacks[i].append(l1)
                    fms = feature_map_size(
                        fms, l1.kernel_size[0], l1.stride[0], l1.padding[0])

            for band, layers in zip(bands, self.layer_stacks):
                for layer in layers:
                    self.add_module(
                        '{band}-{channels}'.format(
                            band=band, channels=layer.out_channels), layer)

            self.total_features = sum(
                [layers[-1].out_channels for layers in self.layer_stacks])
            self.l1 = nn.Linear(
                self.total_features, self.bottleneck, bias=False)
            self.bn1 = nn.BatchNorm1d(self.bottleneck)

        def forward(self, x):
            x = x.view(-1, 1, self.total_bands)

            fms = []
            start_index = 0
            for i, band in enumerate(self.bands):
                stop = start_index + band
                fm = x[..., start_index: stop]
                start_index = stop
                subset = self.layer_stacks[i]
                for s in subset:
                    fm = s(fm)
                fms.append(fm)

            # push the band-wise frequency maps through some linear layers
            flat = torch.cat(fms, dim=1).squeeze()

            x = self.l1(flat)
            x = self.bn1(x)
            x = F.leaky_relu(x, 0.2)
            x = F.dropout(x, 0.2, self.training)

            return x


    class FrequencyDecompositionGenerator(nn.Module):
        def __init__(
                self,
                first_feature_map_size,
                growth_factor,
                bands,
                z_dim):

            super(FrequencyDecompositionGenerator, self).__init__()
            self.first_feature_map_size = first_feature_map_size
            self.growth_factor = growth_factor
            self.bands = bands
            self.z_dim = z_dim
            self.layer_stacks = [[] for _ in bands]

            for i, band in enumerate(bands):
                out_channels = 1
                in_channels = self.first_feature_map_size
                l1 = ConvTranspose1d(
                    in_channels, out_channels, 8, 4, 0, F.tanh, dropout=False)
                self.layer_stacks[i].append(l1)
                fms = feature_map_size(
                    band, l1.kernel_size[0], l1.stride[0], l1.padding[0])
                while fms > 1:
                    out_channels = in_channels
                    in_channels = int(in_channels * self.growth_factor)
                    fms = feature_map_size(
                        fms, 3, 2, 0)
                    in_channels = self.z_dim if fms == 1 else in_channels
                    l1 = ConvTranspose1d(in_channels, out_channels, 3, 2, 0)
                    self.layer_stacks[i].append(l1)

            for band, layers in zip(bands, self.layer_stacks):
                for layer in layers:
                    self.add_module(
                        '{band}-{channels}'.format(
                            band=band, channels=layer.out_channels), layer)

            self.feature_map_sizes = \
                [layers[-1].in_channels for layers in self.layer_stacks]
            self.total_features = sum(self.feature_map_sizes)
            # self.l1 = nn.Linear(self.z_dim, 256, bias=False)
            # self.bn1 = nn.BatchNorm1d(256)
            # self.l2 = nn.Linear(256, self.total_features, bias=False)
            # self.bn2 = nn.BatchNorm1d(self.total_features)

        def forward(self, x):

            # x = x.view(-1, self.z_dim)
            # x = self.l1(x)
            # x = self.bn1(x)
            # x = F.leaky_relu(x, 0.2)
            # x = F.dropout(x, 0.2, self.training)
            #
            # x = self.l2(x)
            # x = self.bn2(x)
            # x = F.leaky_relu(x, 0.2)
            # x = F.dropout(x, 0.2, self.training)
            #
            # x = x.view(-1, self.total_features, 1)

            x = x.view(-1, self.z_dim, 1)

            current = 0
            bands = []
            for i, fms in enumerate(self.feature_map_sizes):
                stop = current + fms
                # fm = x[:, current: stop, :]
                fm = x
                current = stop
                subset = self.layer_stacks[i][::-1]
                for s in subset:
                    fm = s(fm)
                bands.append(fm.view(-1, fm.size()[-1]))

            return torch.cat(bands, dim=1)

except ImportError:
    pass



from frequencydimension import \
    FrequencyDimensionEncoder, FrequencyDimensionDecoder, \
    ExplicitFrequencyDimensionEncoder, ExplicitFrequencyDimensionDecoder
from identitydimension import IdentityDimensionEncoder, IdentityDimensionDecoder
from timedimension import TimeDimensionEncoder, TimeDimensionDecoder


class DimensionEncoder(object):
    encoders = [
        IdentityDimensionEncoder(),
        TimeDimensionEncoder(),
        FrequencyDimensionEncoder(),
        ExplicitFrequencyDimensionEncoder()
    ]

    def __init__(self):
        super(DimensionEncoder, self).__init__()

    def encode(self, o):
        for dim in o:
            for encoder in self.encoders:
                if encoder.matches(dim):
                    yield encoder.encode(dim)
                    break
            else:
                raise NotImplementedError(
                        'No matching strategy for {dim}'.format(**locals()))


class DimensionDecoder(object):
    decoders = [
        IdentityDimensionDecoder(),
        TimeDimensionDecoder(),
        FrequencyDimensionDecoder(),
        ExplicitFrequencyDimensionDecoder()
    ]

    def __init__(self):
        super(DimensionDecoder, self).__init__()

    def decode(self, d):
        for dim in d:
            for decoder in self.decoders:
                if decoder.matches(dim):
                    yield decoder.decode(dim)
                    break
            else:
                raise NotImplementedError(
                        'No matching strategy for {dim}'.format(**locals()))



import unittest2
from loudness import mu_law, inverse_mu_law
import numpy as np


class TestLoudness(unittest2.TestCase):
    def test_can_invert_mu_law(self):
        a = np.random.normal(0, 1, (100, 4))
        adjusted = mu_law(a)
        inverted = inverse_mu_law(adjusted)
        np.testing.assert_allclose(a, inverted)



from __future__ import division
import numpy as np
import unittest2
from frequencyscale import LinearScale, FrequencyBand, Hertz
from weighting import AWeighting
from zounds.timeseries import Seconds, TimeDimension, TimeSlice, SR11025
from zounds.spectral import FrequencyDimension, GeometricScale
from zounds.core import ArrayWithUnits, IdentityDimension


class TimeFrequencyRepresentationTests(unittest2.TestCase):

    def test_can_slice_frequency_dim_with_end_hz(self):
        scale = LinearScale(FrequencyBand(0, 100), 10)
        arr = ArrayWithUnits(
            np.zeros((13, 10)),
            [IdentityDimension(), FrequencyDimension(scale)])
        sliced = arr[:, :Hertz(50)]
        self.assertEqual((13, 5), sliced.shape)
        self.assertEqual(
            FrequencyDimension(LinearScale(FrequencyBand(0, 50), 5)),
            sliced.dimensions[-1])

    def test_can_slice_frequency_dim_with_start_hz(self):
        scale = LinearScale(FrequencyBand(0, 100), 10)
        arr = ArrayWithUnits(
            np.zeros((13, 10)),
            [IdentityDimension(), FrequencyDimension(scale)])
        sliced = arr[:, Hertz(50):]
        self.assertEqual((13, 6), sliced.shape)

    def test_can_slice_frequency_dim_with_start_and_end_hz(self):
        scale = LinearScale(FrequencyBand(0, 100), 10)
        arr = ArrayWithUnits(
            np.zeros((13, 10)),
            [IdentityDimension(), FrequencyDimension(scale)])
        sliced = arr[:, Hertz(20):Hertz(80)]
        self.assertEqual((13, 7), sliced.shape)

    def test_can_slice_frequency_dim_with_negative_start_hz(self):
        scale = LinearScale(FrequencyBand(0, 100), 10)
        arr = ArrayWithUnits(
            np.zeros((13, 10)),
            [IdentityDimension(), FrequencyDimension(scale)])
        sliced = arr[:, -Hertz(20):]
        self.assertEqual((13, 3), sliced.shape)

    def test_can_slice_frequency_dim_with_negative_stop_hz(self):
        scale = LinearScale(FrequencyBand(0, 100), 10)
        arr = ArrayWithUnits(
            np.zeros((13, 10)),
            [IdentityDimension(), FrequencyDimension(scale)])
        sliced = arr[:, :-Hertz(20)]
        self.assertEqual((13, 8), sliced.shape)
        self.assertEqual(
            FrequencyDimension(LinearScale(FrequencyBand(0, 80), 8)),
            sliced.dimensions[-1])

    def test_sliding_window_has_correct_dimensions(self):
        arr = np.random.randint(0, 255, (11025 * 2)).astype(np.int64)
        sr = SR11025()
        awu = ArrayWithUnits(arr, [TimeDimension(*sr)])
        ws = TimeSlice(duration=sr.frequency * 8192)
        ss = TimeSlice(duration=sr.frequency * 4096)
        l, x = awu.sliding_window_with_leftovers(ws, ss)
        self.assertEqual(8192, x.shape[1])

    def test_can_transpose(self):
        sr = SR11025()
        hl = sr.half_lapped()
        scale = GeometricScale(20, sr.nyquist, 0.175, 64)
        td = TimeDimension(frequency=hl.frequency, duration=hl.duration)
        fd = FrequencyDimension(scale)

        arr = ArrayWithUnits(np.zeros((99, 64)), [td, fd])
        transposed = arr.T
        self.assertEqual((64, 99), transposed.shape)
        self.assertEqual(arr.dimensions[0], transposed.dimensions[1])
        self.assertEqual(arr.dimensions[1], transposed.dimensions[0])

    @unittest2.skip('this requires that frequency scales be reversible')
    def test_can_rotate_90_degrees(self):
        sr = SR11025()
        hl = sr.half_lapped()
        scale = GeometricScale(20, sr.nyquist, 0.175, 64)
        td = TimeDimension(frequency=hl.frequency, duration=hl.duration)
        fd = FrequencyDimension(scale)

        arr = ArrayWithUnits(np.zeros((99, 64)), [td, fd])
        rotated = np.rot90(arr)
        self.assertEqual((64, 99), rotated.shape)
        self.assertEqual(arr.dimensions[0], rotated.dimensions[1])
        self.assertEqual(arr.dimensions[1], rotated.dimensions[0])

    def test_can_apply_sliding_window(self):
        sr = SR11025()
        hl = sr.half_lapped()
        scale = GeometricScale(20, sr.nyquist, 0.175, 64)
        td = TimeDimension(frequency=hl.frequency, duration=hl.duration)
        fd = FrequencyDimension(scale)

        arr = ArrayWithUnits(np.zeros((99, 64)), [td, fd])

        ts = TimeSlice(duration=hl.frequency * 64)
        fs = FrequencyBand(0, sr.nyquist)

        windowed = arr.sliding_window((ts, fs))
        self.assertEqual((1, 64, 64), windowed.shape)

    def test_can_iterate_over_time_frequency_representation(self):
        tf = ArrayWithUnits(
            np.ones((10, 10)),
            dimensions=[
                TimeDimension(Seconds(1), Seconds(1)),
                FrequencyDimension(LinearScale(FrequencyBand(0, 1000), 10))
            ])
        rows = [row for row in tf]
        self.assertEqual(10, len(rows))
        for row in rows:
            self.assertIsInstance(row, ArrayWithUnits)
            self.assertIsInstance(row.dimensions[0], FrequencyDimension)
            self.assertEqual((10,), row.shape)

    def test_can_iterate_over_timeseries(self):
        tf = ArrayWithUnits(
            np.ones((10, 10)),
            dimensions=[
                TimeDimension(Seconds(1), Seconds(1)),
                IdentityDimension()
            ])
        rows = [row for row in tf]
        self.assertEqual(10, len(rows))
        for row in rows:
            self.assertIsInstance(row, ArrayWithUnits)
            self.assertIsInstance(row.dimensions[0], IdentityDimension)
            self.assertEqual((10,), row.shape)

    def test_can_iterate_after_packbits(self):
        tf = ArrayWithUnits(
            np.random.binomial(1, 0.5, (10, 256)).astype(np.uint8),
            dimensions=[
                TimeDimension(Seconds(1), Seconds(1)),
                IdentityDimension()
            ])
        tf = tf.packbits(axis=1)
        self.assertIsInstance(tf, ArrayWithUnits)
        self.assertEqual((10, 32), tf.shape)
        self.assertIsInstance(tf.dimensions[0], TimeDimension)
        self.assertIsInstance(tf.dimensions[1], IdentityDimension)
        rows = [row for row in tf]
        self.assertEqual(10, len(rows))
        for row in rows:
            self.assertIsInstance(row, ArrayWithUnits)
            self.assertIsInstance(row.dimensions[0], IdentityDimension)
            self.assertEqual((32,), row.shape)

    def test_can_use_tuple_indices_for_first_dimension(self):
        tf = ArrayWithUnits(
            np.ones((10, 10)),
            dimensions=[
                TimeDimension(Seconds(1), Seconds(1)),
                FrequencyDimension(LinearScale(FrequencyBand(0, 1000), 10))
            ])
        subset = tf[tuple([2, 4, 6]), ...]
        self.assertEqual((3, 10), subset.shape)
        self.assertIsInstance(subset, ArrayWithUnits)
        self.assertIsInstance(subset.dimensions[0], TimeDimension)
        self.assertIsInstance(subset.dimensions[1], FrequencyDimension)

    def test_can_use_tuple_indices_for_first_dimension_id_dim_first(self):
        tf = ArrayWithUnits(
            np.ones((10, 9, 8)),
            dimensions=[
                IdentityDimension(),
                TimeDimension(Seconds(1), Seconds(1)),
                FrequencyDimension(LinearScale(FrequencyBand(0, 1000), 8))
            ])
        subset = tf[tuple([2, 4, 6]), ...]
        self.assertEqual((3, 9, 8), subset.shape)
        self.assertIsInstance(subset, ArrayWithUnits)
        self.assertIsInstance(subset.dimensions[0], IdentityDimension)
        self.assertIsInstance(subset.dimensions[1], TimeDimension)
        self.assertIsInstance(subset.dimensions[2], FrequencyDimension)

    def test_can_access_int_index_and_frequency_band(self):
        tf = ArrayWithUnits(
            np.ones((10, 10)),
            dimensions=[
                TimeDimension(Seconds(1), Seconds(1)),
                FrequencyDimension(LinearScale(FrequencyBand(0, 1000), 10))
            ])
        sliced = tf[0, FrequencyBand(201, 400)]
        self.assertEqual((2,), sliced.shape)
        self.assertIsInstance(sliced.dimensions[0], FrequencyDimension)

    def test_can_access_time_slice_and_int_index(self):
        tf = ArrayWithUnits(
            np.ones((10, 10)),
            dimensions=[
                TimeDimension(Seconds(1), Seconds(1)),
                FrequencyDimension(LinearScale(FrequencyBand(0, 1000), 10))
            ])
        sliced = tf[TimeSlice(start=Seconds(1), duration=Seconds(2)), 0]
        self.assertEqual((2,), sliced.shape)
        self.assertIsInstance(sliced.dimensions[0], TimeDimension)

    def test_can_add_axis_at_end(self):
        _id = IdentityDimension()
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.ones((3, 30, 100)), [_id, td, fd])
        tf2 = tf[..., None]
        self.assertEqual(4, tf2.ndim)
        self.assertIsInstance(tf2.dimensions[0], IdentityDimension)
        self.assertIsInstance(tf2.dimensions[1], TimeDimension)
        self.assertIsInstance(tf2.dimensions[2], FrequencyDimension)
        self.assertIsInstance(tf2.dimensions[3], IdentityDimension)

    def test_sum_along_frequency_axis(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.ones((30, 100)), [td, fd])
        result = tf.sum(axis=1)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertEqual((30,), result.shape)
        self.assertIsInstance(result.dimensions[0], TimeDimension)

    def test_can_use_negative_axis_indices(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.ones((30, 100)), [td, fd])
        result = tf.sum(axis=-1)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertEqual((30,), result.shape)
        self.assertIsInstance(result.dimensions[0], TimeDimension)

    def test_can_use_keepdims_with_sum(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.ones((30, 100)), [td, fd])
        result = tf.sum(axis=-1, keepdims=True)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(2, len(result.dimensions))
        self.assertEqual((30, 1), result.shape)
        self.assertIsInstance(result.dimensions[0], TimeDimension)
        self.assertIsInstance(result.dimensions[1], IdentityDimension)

    def test_can_use_negative_axis_indices_max(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.ones((30, 100)), [td, fd])
        result = tf.max(axis=-1)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(1, len(result.dimensions))
        self.assertEqual((30,), result.shape)
        self.assertIsInstance(result.dimensions[0], TimeDimension)

    def test_from_example(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.ones((30, 100)), [td, fd])
        from_example = ArrayWithUnits.from_example(np.ones((30, 100)), tf)
        self.assertEqual(tf.shape, from_example.shape)
        self.assertItemsEqual(tf.dimensions, from_example.dimensions)

    def test_can_multiply_by_frequency_weighting_linear_scale(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)

        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)

        tf = ArrayWithUnits(np.ones((30, 100)), [td, fd])
        result = tf * AWeighting()
        self.assertIsInstance(result, ArrayWithUnits)
        peak_frequency_band = FrequencyBand(9000, 11000)
        lower_band = FrequencyBand(100, 300)
        peak_slice = np.abs(result[:, peak_frequency_band]).max()
        lower_slice = np.abs(result[:, lower_band]).max()
        self.assertGreater(peak_slice, lower_slice)

    def test_can_multiply_by_frequency_weighting_log_scale(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = GeometricScale(20, 22050, 0.01, 100)

        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)

        tf = ArrayWithUnits(np.ones((30, 100)), [td, fd])

        result = tf * AWeighting()
        self.assertIsInstance(result, ArrayWithUnits)
        peak_frequency_band = FrequencyBand(9000, 11000)
        lower_band = FrequencyBand(100, 300)
        peak_slice = np.abs(result[:, peak_frequency_band]).max()
        lower_slice = np.abs(result[:, lower_band]).max()
        self.assertGreater(peak_slice, lower_slice)

    def test_can_multiply_by_array(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)
        tf = ArrayWithUnits(np.ones((30, 100)), [td, fd])
        result = tf * np.ones(100)
        self.assertIsInstance(result, ArrayWithUnits)
        np.testing.assert_allclose(tf, result)

    def test_can_use_list_of_integers_as_index(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)

        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)

        tf = ArrayWithUnits(np.zeros((30, 100)), [td, fd])

        indexed = tf[[0, 10, 14]]
        self.assertEqual((3, 100), indexed.shape)
        self.assertIsInstance(indexed, ArrayWithUnits)
        self.assertIsInstance(indexed.dimensions[0], IdentityDimension)
        self.assertIsInstance(indexed.dimensions[1], FrequencyDimension)

    def test_can_construct_instance(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)
        tf = ArrayWithUnits(np.zeros((30, 100)), [td, fd])
        self.assertIsInstance(tf, ArrayWithUnits)

    def test_raises_if_scale_length_does_not_match_frequency_dimension(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 1000)

        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)

        self.assertRaises(
            ValueError,
            lambda: ArrayWithUnits(np.ones((30, 100)), [td, fd]))

    def test_can_slice_frequency_dimension_with_integer_indices(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)
        tf = ArrayWithUnits(np.zeros((30, 100)), [td, fd])
        sliced = tf[:, 10: 20]
        self.assertEqual((30, 10), sliced.shape)
        self.assertIsInstance(sliced, ArrayWithUnits)

    def test_can_slice_frequency_dimensions_with_frequency_band(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)
        tf = ArrayWithUnits(np.zeros((30, 100)), [td, fd])
        bands = list(scale)
        sliced = tf[:, bands[0]]
        self.assertEqual((30, 1), sliced.shape)
        self.assertIsInstance(sliced, ArrayWithUnits)

    def test_can_slice_freq_dimension_with_freq_band_spanning_bins(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)
        tf = ArrayWithUnits(np.zeros((30, 100)), [td, fd])
        bands = list(scale)
        wide_band = FrequencyBand(bands[0].start_hz, bands[9].stop_hz)
        sliced = tf[:, wide_band]
        self.assertEqual((30, 10), sliced.shape)
        self.assertIsInstance(sliced, ArrayWithUnits)

    def test_scale_is_modified_after_slice(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        td = TimeDimension(frequency, duration)
        fd = FrequencyDimension(scale)
        tf = ArrayWithUnits(np.zeros((30, 100)), [td, fd])
        bands = list(scale)
        wide_band = FrequencyBand(bands[0].start_hz, bands[9].stop_hz)
        sliced = tf[:, wide_band]
        self.assertEqual((30, 10), sliced.shape)
        self.assertIsInstance(sliced, ArrayWithUnits)
        self.assertLess(sliced.dimensions[1].scale.stop_hz, scale.stop_hz)
        self.assertEqual(10, sliced.dimensions[1].scale.n_bands)

    def test_ellipsis(self):
        scale = LinearScale(FrequencyBand(0, 10000), 100)
        arr = ArrayWithUnits(
            np.zeros((10, 3, 100)),
            [IdentityDimension(),
             TimeDimension(Seconds(1)),
             FrequencyDimension(scale)])
        sliced = arr[..., FrequencyBand(1000, 5000)]
        self.assertEqual((10, 3, 41), sliced.shape)
        self.assertIsInstance(sliced.dimensions[0], IdentityDimension)
        self.assertIsInstance(sliced.dimensions[1], TimeDimension)
        self.assertIsInstance(sliced.dimensions[2], FrequencyDimension)



import unittest2
from tfrepresentation import FrequencyDimension, ExplicitFrequencyDimension
from frequencyscale import FrequencyBand, LinearScale, GeometricScale


class FrequencyDimensionTests(unittest2.TestCase):
    def test_equal(self):
        fd1 = FrequencyDimension(LinearScale(FrequencyBand(20, 10000), 100))
        fd2 = FrequencyDimension(LinearScale(FrequencyBand(20, 10000), 100))
        self.assertEqual(fd1, fd2)

    def test_not_equal(self):
        fd1 = FrequencyDimension(LinearScale(FrequencyBand(20, 10000), 100))
        fd2 = FrequencyDimension(GeometricScale(20, 10000, 0.01, 100))
        self.assertNotEqual(fd1, fd2)


class ExplicitFrequencyDimensionTests(unittest2.TestCase):
    def test_equal(self):
        scale1 = GeometricScale(20, 5000, 0.02, 3)
        slices1 = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim1 = ExplicitFrequencyDimension(scale1, slices1)

        scale2 = GeometricScale(20, 5000, 0.02, 3)
        slices2 = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim2 = ExplicitFrequencyDimension(scale2, slices2)

        self.assertEqual(dim1, dim2)

    def test_not_equal_due_to_slices(self):
        scale1 = GeometricScale(20, 5000, 0.02, 3)
        slices1 = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim1 = ExplicitFrequencyDimension(scale1, slices1)

        scale2 = GeometricScale(20, 5000, 0.02, 3)
        slices2 = [slice(0, 10), slice(10, 100), slice(100, 1001)]
        dim2 = ExplicitFrequencyDimension(scale2, slices2)

        self.assertNotEqual(dim1, dim2)

    def test_not_equal_due_to_scales(self):
        scale1 = GeometricScale(20, 5000, 0.02, 3)
        slices1 = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim1 = ExplicitFrequencyDimension(scale1, slices1)

        scale2 = GeometricScale(20, 5000, 0.02, 3)
        slices2 = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim2 = ExplicitFrequencyDimension(scale2, slices2)

        self.assertNotEqual(dim1, dim2)

    def test_raises_when_scale_and_slices_are_different_sizes(self):
        scale1 = GeometricScale(20, 5000, 0.02, 3)
        slices1 = [slice(0, 10), slice(10, 100)]
        self.assertRaises(
            ValueError, lambda: ExplicitFrequencyDimension(scale1, slices1))

    def test_metaslice(self):
        scale1 = GeometricScale(20, 5000, 0.02, 3)
        bands = list(scale1)
        slices1 = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim1 = ExplicitFrequencyDimension(scale1, slices1)
        dim2 = dim1.metaslice(FrequencyBand(15, 1000), 2)
        self.assertEqual(bands[:2], list(dim2.scale)[:2])
        self.assertEqual(slices1[:2], dim2.slices[:2])

    def test_metaslice_exact_matching_band(self):
        scale = GeometricScale(20, 5000, 0.05, 120)
        # the values of the slices don't matter for this test
        slices = [slice(0, 10) for _ in xrange(len(scale))]
        dim = ExplicitFrequencyDimension(scale, slices)
        dim2 = dim.metaslice(scale[0], 1)
        self.assertEqual(1, len(dim2.scale))
        self.assertEqual(1, len(dim2.slices))
        self.assertEqual(dim.scale[0], dim2.scale[0])
        self.assertEqual(dim.slices[0], dim2.slices[0])

    def test_metaslice_fuzzy_matching_band(self):
        scale = GeometricScale(20, 5000, 0.05, 120)
        # the values of the slices don't matter for this test
        slices = [slice(0, 10) for _ in xrange(len(scale))]
        dim = ExplicitFrequencyDimension(scale, slices)
        first_band = scale[0]
        band = FrequencyBand(first_band.start_hz, first_band.stop_hz + 1)
        dim2 = dim.metaslice(band, 3)
        self.assertEqual(3, len(dim2.scale))
        self.assertEqual(3, len(dim2.slices))
        self.assertEqual(dim.scale[0], dim2.scale[0])
        self.assertEqual(dim.scale[1], dim2.scale[1])
        self.assertEqual(dim.scale[2], dim2.scale[2])
        self.assertEqual(dim.slices[0], dim2.slices[0])
        self.assertEqual(dim.slices[1], dim2.slices[1])
        self.assertEqual(dim.slices[2], dim2.slices[2])

    def test_can_get_slice_when_perfectly_corresponds_to_band(self):
        scale1 = GeometricScale(20, 5000, 0.02, 3)
        bands = list(scale1)
        slices1 = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim1 = ExplicitFrequencyDimension(scale1, slices1)
        self.assertEqual(slices1[1], dim1.integer_based_slice(bands[1]))

    def test_can_get_slice_with_overlap(self):
        scale1 = GeometricScale(20, 5000, 0.02, 3)
        bands = list(scale1)
        slices1 = [slice(0, 10), slice(5, 100), slice(50, 1000)]
        dim1 = ExplicitFrequencyDimension(scale1, slices1)
        self.assertEqual(slices1[1], dim1.integer_based_slice(bands[1]))

    def test_is_valid_when_size_corresponds_to_last_slice_end(self):
        scale1 = GeometricScale(20, 5000, 0.02, 3)
        slices1 = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim1 = ExplicitFrequencyDimension(scale1, slices1)
        self.assertTrue(dim1.validate(1000))

    def test_is_not_valid_when_size_does_not_correspond_to_last_slice(self):
        scale1 = GeometricScale(20, 5000, 0.02, 3)
        slices1 = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim1 = ExplicitFrequencyDimension(scale1, slices1)
        self.assertFalse(dim1.validate(2000))



import os
import hashlib
import requests


class DataSetCache(object):
    def __init__(self, path, dataset):
        super(DataSetCache, self).__init__()
        self.dataset = dataset
        self.path = path
        if not os.path.exists(path):
            os.makedirs(path)

    def __iter__(self):
        for request in self.dataset:
            h = hashlib.md5(request.url).hexdigest()
            p = os.path.join(self.path, h)

            if not os.path.exists(p):
                s = requests.Session()
                prepped = request.prepare()
                resp = s.send(prepped, stream=True)

                resp.raise_for_status()

                with open(p, 'wb') as f:
                    for chunk in resp.iter_content(1024):
                        f.write(chunk)

            yield p



import unittest2
import numpy as np
from weighting import AWeighting
from frequencyscale import LinearScale, FrequencyBand, GeometricScale
from tfrepresentation import FrequencyDimension
from frequencyadaptive import FrequencyAdaptive
from zounds.timeseries import Seconds, TimeDimension, Milliseconds
from zounds.core import ArrayWithUnits, IdentityDimension


class WeightingTests(unittest2.TestCase):
    def test_cannot_multiply_when_array_does_not_have_expected_dimensions(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        tf = ArrayWithUnits(np.ones((90, 100)), [td, IdentityDimension()])
        weighting = AWeighting()
        self.assertRaises(ValueError, lambda: tf * weighting)

    def test_can_get_weights_from_tf_representation(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.ones((90, 100)), [td, fd])
        weighting = AWeighting()
        weights = weighting.weights(tf)
        self.assertEqual((100,), weights.shape)

    def test_can_get_weights_from_scale(self):
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        weighting = AWeighting()
        weights = weighting.weights(scale)
        self.assertEqual((100,), weights.shape)

    def test_can_apply_a_weighting_to_time_frequency_representation(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.ones((90, 100)), [td, fd])
        weighting = AWeighting()
        result = tf * weighting
        self.assertGreater(result[0, -1], result[0, 0])

    def test_can_apply_a_weighting_to_frequency_adaptive_representation(self):
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Milliseconds(500))
        scale = GeometricScale(20, 5000, 0.05, 120)
        arrs = [np.ones((10, x)) for x in xrange(1, 121)]
        fa = FrequencyAdaptive(arrs, td, scale)
        weighting = AWeighting()
        result = fa * weighting
        self.assertGreater(
            result[:, scale[-1]].sum(), result[:, scale[0]].sum())

    def test_can_invert_frequency_weighting(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.random.random_sample((90, 100)), [td, fd])
        weighted = tf * AWeighting()
        inverted = weighted / AWeighting()
        np.testing.assert_allclose(tf, inverted)

    def test_can_invert_frequency_weighting_for_adaptive_representation(self):
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Milliseconds(500))
        scale = GeometricScale(20, 5000, 0.05, 120)
        arrs = [np.random.random_sample((10, x)) for x in xrange(1, 121)]
        fa = FrequencyAdaptive(arrs, td, scale)
        weighting = AWeighting()
        result = fa * weighting
        inverted = result / AWeighting()
        np.testing.assert_allclose(fa, inverted)

    def test_can_apply_weighting_to_explicit_frequency_dimension(self):
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Milliseconds(500))
        scale = GeometricScale(20, 5000, 0.05, 120)
        arrs = [np.ones((10, x)) for x in xrange(1, 121)]
        fa = FrequencyAdaptive(arrs, td, scale)
        fa2 = ArrayWithUnits(fa, fa.dimensions)
        weighting = AWeighting()
        result = fa2 * weighting
        self.assertGreater(
            result[:, scale[-1]].sum(), result[:, scale[0]].sum())

    def test_can_invert_weighting_for_explicit_frequency_dimension(self):
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Milliseconds(500))
        scale = GeometricScale(20, 5000, 0.05, 120)
        arrs = [np.ones((10, x)) for x in xrange(1, 121)]
        fa = FrequencyAdaptive(arrs, td, scale)
        fa2 = ArrayWithUnits(fa, fa.dimensions)
        weighting = AWeighting()
        result = fa2 * weighting
        inverted = result / AWeighting()
        np.testing.assert_allclose(fa, inverted)



from featureflow import Node, Aggregator
import os
from soundfile import SoundFile
import requests
import json
import io
from urlparse import urlparse
import featureflow as ff


class AudioMetaData(object):
    """
    Encapsulates metadata about a source audio file, including things like
    text descriptions and licensing information.

    Args:
        uri (requests.Request or str): uri may be either a string representing
            a network resource or a local file, or a :class:`requests.Request`
            instance
        samplerate (int): the samplerate of the source audio
        channels (int): the number of channels of the source audio
        licensing (str): The licensing agreement (if any) that applies to the
            source audio
        description (str): a text description of the source audio
        tags (str): text tags that apply to the source audio
        kwargs (dict): other arbitrary properties about the source audio

    Raises:
        ValueError: when `uri` is not provided

    See Also:
        :class:`zounds.datasets.FreeSoundSearch`
        :class:`zounds.datasets.InternetArchive`
        :class:`zounds.datasets.PhatDrumLoops`
    """

    def __init__(
            self,
            uri=None,
            samplerate=None,
            channels=None,
            licensing=None,
            description=None,
            tags=None,
            **kwargs):
        super(AudioMetaData, self).__init__()

        if not uri:
            raise ValueError('You must at least supply a uri')

        self.uri = uri
        self.samplerate = samplerate
        self.channels = channels
        self.licensing = licensing
        self.description = description
        self.tags = tags

        for k, v in kwargs.iteritems():
            setattr(self, k, v)

    @property
    def request(self):
        if hasattr(self.uri, 'url'):
            return self.uri

    def __eq__(self, other):
        return self.uri == other.uri \
               and self.samplerate == other.samplerate \
               and self.channels == other.channels \
               and self.licensing == other.licensing \
               and self.description == other.description \
               and self.tags == other.tags

    def __repr__(self):
        return self.__dict__.__str__()

    def __str__(self):
        return self.__repr__()


class AudioMetaDataEncoder(Aggregator, Node):
    content_type = 'application/json'

    def __init__(self, needs=None):
        super(AudioMetaDataEncoder, self).__init__(needs=needs)

    def _uri(self, uri):
        if isinstance(uri, requests.Request):
            return uri.url
        elif isinstance(uri, io.BytesIO) or isinstance(uri, ff.ZipWrapper):
            return None
        else:
            return uri

    def _process(self, data):
        d = dict(data.__dict__)
        d['uri'] = self._uri(data.uri)
        yield json.dumps(d)


class MetaData(Node):
    def __init__(self, needs=None):
        super(MetaData, self).__init__(needs=needs)

    @staticmethod
    def _is_url(s):
        if not isinstance(s, str):
            return False
        parsed = urlparse(s)
        return parsed.scheme and parsed.netloc

    @staticmethod
    def _is_local_file(s):
        try:
            return os.path.exists(s)
        except TypeError:
            return False

    @staticmethod
    def _is_file(s):
        try:
            s.tell()
            return True
        except AttributeError:
            return False

    def _process(self, data):
        if isinstance(data, AudioMetaData):
            yield data
        elif self._is_url(data):
            req = requests.Request(
                method='GET',
                url=data,
                headers={'Range': 'bytes=0-'})
            yield AudioMetaData(uri=req)
        elif isinstance(data, requests.Request):
            if 'range' not in data.headers:
                data.headers['range'] = 'bytes=0-'
            yield AudioMetaData(uri=data)
        elif self._is_local_file(data) or self._is_file(data):
            sf = SoundFile(data)
            yield AudioMetaData(
                uri=data,
                samplerate=sf.samplerate,
                channels=sf.channels)
        else:
            yield AudioMetaData(uri=data)



from __future__ import division

import numpy as np
from scipy.fftpack import dct, idct
from scipy.signal import resample

from zounds.core import ArrayWithUnits, IdentityDimension
from zounds.spectral import DCTIV, LinearScale
from zounds.spectral import FrequencyDimension
from zounds.spectral.sliding_window import \
    IdentityWindowingFunc, OggVorbisWindowingFunc
from zounds.timeseries import \
    nearest_audio_sample_rate, Seconds, AudioSamples, TimeDimension


class ShortTimeTransformSynthesizer(object):
    def __init__(self):
        super(ShortTimeTransformSynthesizer, self).__init__()

    def _transform(self, frames):
        return frames

    def _windowing_function(self):
        return IdentityWindowingFunc()

    def _overlap_add(self, frames):
        time_dim = frames.dimensions[0]
        sample_freq = time_dim.duration / frames.shape[-1]
        windowsize = int(np.round(time_dim.duration / sample_freq))
        hopsize = int(np.round(time_dim.frequency / sample_freq))

        # create an empty array of audio samples
        arr = np.zeros(int(time_dim.end / sample_freq))
        for i, f in enumerate(frames):
            start = i * hopsize
            stop = start + windowsize
            l = len(arr[start:stop])
            arr[start:stop] += (self._windowing_function() * f[:l])

        sr = nearest_audio_sample_rate(Seconds(1) / sample_freq)
        return AudioSamples(arr, sr)

    def synthesize(self, frames):
        audio = self._transform(frames)
        ts = ArrayWithUnits(audio, [frames.dimensions[0], IdentityDimension()])
        return self._overlap_add(ts)


class WindowedAudioSynthesizer(ShortTimeTransformSynthesizer):
    def __init__(self):
        super(WindowedAudioSynthesizer, self).__init__()


class FFTSynthesizer(ShortTimeTransformSynthesizer):
    """
    Inverts the short-time fourier transform, e.g. the output of the
    :class:`~zounds.spectral.FFT` processing node.

    Here's an example that extracts a short-time fourier transform, and then
    inverts it.

    .. code:: python

        import zounds

        STFT = zounds.stft(
            resample_to=zounds.SR11025(),
            store_fft=True)


        @zounds.simple_in_memory_settings
        class Sound(STFT):
            pass

        # produce some additive sine waves
        sine_synth = zounds.SineSynthesizer(zounds.SR22050())
        samples = sine_synth.synthesize(
            zounds.Seconds(4), freqs_in_hz=[220, 400, 880])

        # process the sound, including a short-time fourier transform feature
        _id = Sound.process(meta=samples.encode())
        snd = Sound(_id)

        # invert the frequency-domain feature to reover the original audio
        fft_synth = zounds.FFTSynthesizer()
        recon = fft_synth.synthesize(snd.fft)
        print recon.__class__  #  AudioSamples instance with reconstructed audio

    See Also:
        :class:`~zounds.spectral.FFT`
    """

    def __init__(self):
        super(FFTSynthesizer, self).__init__()

    def _windowing_function(self):
        return OggVorbisWindowingFunc()

    def _transform(self, frames):
        return np.fft.fftpack.irfft(frames, norm='ortho')


class DCTSynthesizer(ShortTimeTransformSynthesizer):
    """
    Inverts the short-time discrete cosine transform (type II), e.g., the output
    of the :class:`~zounds.spectral.DCT` processing node

    Here's an example that extracts a short-time discrete cosine transform, and
    then inverts it.

    .. code:: python

        import zounds

        Resampled = zounds.resampled(resample_to=zounds.SR11025())


        @zounds.simple_in_memory_settings
        class Sound(Resampled):
            windowed = zounds.ArrayWithUnitsFeature(
                zounds.SlidingWindow,
                needs=Resampled.resampled,
                wscheme=zounds.HalfLapped(),
                wfunc=zounds.OggVorbisWindowingFunc(),
                store=False)

            dct = zounds.ArrayWithUnitsFeature(
                zounds.DCT,
                needs=windowed,
                store=True)

        # produce some additive sine waves
        sine_synth = zounds.SineSynthesizer(zounds.SR22050())
        samples = sine_synth.synthesize(
            zounds.Seconds(4), freqs_in_hz=[220, 400, 880])

        # process the sound, including a short-time fourier transform feature
        _id = Sound.process(meta=samples.encode())
        snd = Sound(_id)

        # invert the frequency-domain feature to reover the original audio
        dct_synth = zounds.DCTSynthesizer()
        recon = dct_synth.synthesize(snd.dct)
        print recon.__class__  # AudioSamples instance with reconstructed audio

    See Also:
        :class:`~zounds.spectral.DCT`
    """

    def __init__(self, windowing_func=IdentityWindowingFunc()):
        super(DCTSynthesizer, self).__init__()
        self.windowing_func = windowing_func

    def _windowing_function(self):
        return self.windowing_func

    def _transform(self, frames):
        return idct(frames, norm='ortho')


class DCTIVSynthesizer(ShortTimeTransformSynthesizer):
    """
    Inverts the short-time discrete cosine transform (type IV), e.g., the output
    of the :class:`~zounds.spectral.DCTIV` processing node.

    Here's an example that extracts a short-time DCT-IV transform, and inverts
    it.

    .. code:: python

        import zounds

        Resampled = zounds.resampled(resample_to=zounds.SR11025())


        @zounds.simple_in_memory_settings
        class Sound(Resampled):
            windowed = zounds.ArrayWithUnitsFeature(
                zounds.SlidingWindow,
                needs=Resampled.resampled,
                wscheme=zounds.HalfLapped(),
                wfunc=zounds.OggVorbisWindowingFunc(),
                store=False)

            dct = zounds.ArrayWithUnitsFeature(
                zounds.DCTIV,
                needs=windowed,
                store=True)

        # produce some additive sine waves
        sine_synth = zounds.SineSynthesizer(zounds.SR22050())
        samples = sine_synth.synthesize(
            zounds.Seconds(4), freqs_in_hz=[220, 400, 880])

        # process the sound, including a short-time fourier transform feature
        _id = Sound.process(meta=samples.encode())
        snd = Sound(_id)

        # invert the frequency-domain feature to reover the original audio
        dct_synth = zounds.DCTIVSynthesizer()
        recon = dct_synth.synthesize(snd.dct)
        print recon.__class__  # AudioSamples instance with reconstructed audio

    See Also:
        :class:`~zounds.spectral.DCTIV`
    """

    def __init__(self, windowing_func=IdentityWindowingFunc()):
        super(DCTIVSynthesizer, self).__init__()
        self.windowing_func = windowing_func

    def _windowing_function(self):
        return self.windowing_func

    def _transform(self, frames):
        return list(DCTIV()._process(frames))[0]


class MDCTSynthesizer(ShortTimeTransformSynthesizer):
    """
    Inverts the modified discrete cosine transform, e.g., the output of the
    :class:`~zounds.spectral.MDCT` processing node.

    Here's an example that extracts a short-time MDCT transform, and inverts
    it.

    .. code:: python

        import zounds

        Resampled = zounds.resampled(resample_to=zounds.SR11025())


        @zounds.simple_in_memory_settings
        class Sound(Resampled):
            windowed = zounds.ArrayWithUnitsFeature(
                zounds.SlidingWindow,
                needs=Resampled.resampled,
                wscheme=zounds.HalfLapped(),
                wfunc=zounds.OggVorbisWindowingFunc(),
                store=False)

            mdct = zounds.ArrayWithUnitsFeature(
                zounds.MDCT,
                needs=windowed,
                store=True)

        # produce some additive sine waves
        sine_synth = zounds.SineSynthesizer(zounds.SR22050())
        samples = sine_synth.synthesize(
            zounds.Seconds(4), freqs_in_hz=[220, 400, 880])

        # process the sound, including a short-time fourier transform feature
        _id = Sound.process(meta=samples.encode())
        snd = Sound(_id)

        # invert the frequency-domain feature to reover the original audio
        mdct_synth = zounds.MDCTSynthesizer()
        recon = mdct_synth.synthesize(snd.mdct)
        print recon.__class__  # AudioSamples instance with reconstructed audio

    See Also:
        :class:`~zounds.spectral.MDCT`
    """

    def __init__(self):
        super(MDCTSynthesizer, self).__init__()

    def _windowing_function(self):
        return OggVorbisWindowingFunc()

    def _transform(self, frames):
        l = frames.shape[1]
        t = np.arange(0, 2 * l)
        f = np.arange(0, l)
        cpi = -1j * np.pi
        a = frames * np.exp(cpi * (f + 0.5) * (l + 1) / 2 / l)
        b = np.fft.fft(a, 2 * l)
        return np.sqrt(2 / l) * np.real(b * np.exp(cpi * t / 2 / l))


class FrequencyDecompositionSynthesizer(object):
    def __init__(self, samplerate, output_size):
        super(FrequencyDecompositionSynthesizer, self).__init__()
        self.output_size = output_size
        self.samplerate = samplerate

    def synthesize(self, x, bands=None):
        output = ArrayWithUnits(
            np.zeros((len(x), self.output_size)),
            dimensions=[x.time_dimension, TimeDimension(*self.samplerate)])

        for i, band in enumerate(x.scale):
            if bands and i not in bands:
                continue
            output += resample(x[:, band], self.output_size, axis=-1)

        return output


class BaseFrequencyAdaptiveSynthesizer(object):
    def __init__(
            self,
            scale,
            band_transform,
            short_time_synth,
            samplerate,
            coeffs_dtype,
            scale_slices_always_even):
        super(BaseFrequencyAdaptiveSynthesizer, self).__init__()
        self.scale_slices_always_even = scale_slices_always_even
        self.coeffs_dtype = coeffs_dtype
        self.scale = scale
        self.samplerate = samplerate
        self.short_time_synth = short_time_synth
        self.band_transform = band_transform

    def _n_linear_scale_bands(self, frequency_adaptive_coeffs):
        raise NotImplementedError()

    def synthesize(self, freq_adaptive_coeffs):
        fac = freq_adaptive_coeffs

        linear_scale = LinearScale.from_sample_rate(
            self.samplerate,
            self._n_linear_scale_bands(fac),
            always_even=self.scale_slices_always_even)

        frequency_dimension = FrequencyDimension(linear_scale)

        coeffs = ArrayWithUnits(
            np.zeros((len(fac), linear_scale.n_bands), dtype=self.coeffs_dtype),
            dimensions=[fac.dimensions[0], frequency_dimension])

        for band in self.scale:
            coeffs[:, band] += self.band_transform(fac[:, band], norm='ortho')

        return self.short_time_synth.synthesize(coeffs)


class FrequencyAdaptiveDCTSynthesizer(BaseFrequencyAdaptiveSynthesizer):
    """
    Invert a frequency-adaptive transform, e.g., one produced by the
    :class:`zounds.spectral.FrequencyAdaptiveTransform` processing node which
    has used a discrete cosine transform in its `transform` parameter.

    Args:
        scale (FrequencyScale): The scale used to produce the frequency-adaptive
            transform
        samplerate (SampleRate): The audio samplerate of the audio that was
            originally transformed

    Here's an example of how you might first extract a frequency-adaptive
    representation, and then invert it:

    .. code:: python

        import zounds
        import scipy
        import numpy as np

        samplerate = zounds.SR11025()
        Resampled = zounds.resampled(resample_to=samplerate)

        scale = zounds.GeometricScale(
            100, 5000, bandwidth_ratio=0.089, n_bands=100)
        scale.ensure_overlap_ratio(0.5)


        @zounds.simple_in_memory_settings
        class Sound(Resampled):
            long_windowed = zounds.ArrayWithUnitsFeature(
                zounds.SlidingWindow,
                wscheme=zounds.SampleRate(
                    frequency=zounds.Milliseconds(500),
                    duration=zounds.Seconds(1)),
                wfunc=zounds.OggVorbisWindowingFunc(),
                needs=Resampled.resampled)

            dct = zounds.ArrayWithUnitsFeature(
                zounds.DCT,
                scale_always_even=True,
                needs=long_windowed)

            freq_adaptive = zounds.FrequencyAdaptiveFeature(
                zounds.FrequencyAdaptiveTransform,
                transform=scipy.fftpack.idct,
                window_func=np.hanning,
                scale=scale,
                needs=dct,
                store=True)


        # produce some additive sine waves
        sine_synth = zounds.SineSynthesizer(zounds.SR22050())
        samples = sine_synth.synthesize(
            zounds.Seconds(10), freqs_in_hz=[220, 440, 880])

        # process the sound, including a short-time fourier transform feature
        _id = Sound.process(meta=samples.encode())
        snd = Sound(_id)

        # invert the sound
        synth = zounds.FrequencyAdaptiveDCTSynthesizer(scale, samplerate)
        recon = synth.synthesize(snd.freq_adaptive)
        print recon  # AudioSamples instance with the reconstructed sound

    See Also:
        :class:`~zounds.spectral.DCT`
        :class:`~zounds.spectral.FrequencyAdaptive`
        :class:`~zounds.spectral.FrequencyAdaptiveTransform`
    """

    def __init__(self, scale, samplerate):
        super(FrequencyAdaptiveDCTSynthesizer, self).__init__(
            scale,
            dct,
            DCTSynthesizer(),
            samplerate,
            np.float64,
            scale_slices_always_even=True)

    def _n_linear_scale_bands(self, frequency_adaptive_coeffs):
        fac = frequency_adaptive_coeffs.dimensions[0]
        return int(fac.duration / self.samplerate.frequency)


class FrequencyAdaptiveFFTSynthesizer(BaseFrequencyAdaptiveSynthesizer):
    """
    Invert a frequency-adaptive transform, e.g., one produced by the
    :class:`zounds.spectral.FrequencyAdaptiveTransform` processing node which
    has used a fast fouriter transform in its `transform` parameter.

    Args:
        scale (FrequencyScale): The scale used to produce the frequency-adaptive
            transform
        samplerate (SampleRate): The audio samplerate of the audio that was
            originally transformed

    Here's an example of how you might first extract a frequency-adaptive
    representation, and then invert it:

    .. code:: python

        import zounds
        import numpy as np

        samplerate = zounds.SR11025()
        Resampled = zounds.resampled(resample_to=samplerate)

        scale = zounds.GeometricScale(100, 5000, bandwidth_ratio=0.089, n_bands=100)
        scale.ensure_overlap_ratio(0.5)


        @zounds.simple_in_memory_settings
        class Sound(Resampled):
            long_windowed = zounds.ArrayWithUnitsFeature(
                zounds.SlidingWindow,
                wscheme=zounds.SampleRate(
                    frequency=zounds.Milliseconds(500),
                    duration=zounds.Seconds(1)),
                wfunc=zounds.OggVorbisWindowingFunc(),
                needs=Resampled.resampled)

            fft = zounds.ArrayWithUnitsFeature(
                zounds.FFT,
                needs=long_windowed)

            freq_adaptive = zounds.FrequencyAdaptiveFeature(
                zounds.FrequencyAdaptiveTransform,
                transform=np.fft.irfft,
                window_func=np.hanning,
                scale=scale,
                needs=fft,
                store=True)


        # produce some additive sine waves
        sine_synth = zounds.SineSynthesizer(zounds.SR22050())
        samples = sine_synth.synthesize(
            zounds.Seconds(10), freqs_in_hz=[220, 440, 880])

        # process the sound, including a short-time fourier transform feature
        _id = Sound.process(meta=samples.encode())
        snd = Sound(_id)

        # invert the sound
        synth = zounds.FrequencyAdaptiveFFTSynthesizer(scale, samplerate)
        recon = synth.synthesize(snd.freq_adaptive)
        print recon  # AudioSamples instance with the reconstructed sound

    See Also:
        :class:`~zounds.spectral.FFT`
        :class:`~zounds.spectral.FrequencyAdaptive`
        :class:`~zounds.spectral.FrequencyAdaptiveTransform`
    """

    def __init__(self, scale, samplerate):
        super(FrequencyAdaptiveFFTSynthesizer, self).__init__(
            scale,
            np.fft.rfft,
            FFTSynthesizer(),
            samplerate,
            np.complex128,
            scale_slices_always_even=False)

    def _n_linear_scale_bands(self, frequency_adaptive_coeffs):
        # https://docs.scipy.org/doc/numpy/reference/generated/numpy.fft.rfft.html#numpy.fft.rfft
        fac = frequency_adaptive_coeffs.dimensions[0]
        raw_samples = int(fac.duration / self.samplerate.frequency)
        return int(raw_samples // 2) + 1


class SineSynthesizer(object):
    """
    Synthesize sine waves

    Args:
        samplerate (Samplerate): the samplerate at which the sine waves should
            be synthesized

    Examples:
        >>> import zounds
        >>> synth = zounds.SineSynthesizer(zounds.SR22050())
        >>> samples = synth.synthesize( \
            zounds.Seconds(1), freqs_in_hz=[220., 440.])
        >>> samples
        AudioSamples([ 0.        ,  0.09384942,  0.18659419, ..., -0.27714552,
               -0.18659419, -0.09384942])
        >>> len(samples)
        22050


    See Also:
        :class:`TickSynthesizer`
        :class:`NoiseSynthesizer`
        :class:`SilenceSynthesizer`
    """

    def __init__(self, samplerate):
        super(SineSynthesizer, self).__init__()
        self.samplerate = samplerate

    def synthesize(self, duration, freqs_in_hz=[440.]):
        """
        Synthesize one or more sine waves

        Args:
            duration (numpy.timdelta64): The duration of the sound to be
                synthesized
            freqs_in_hz (list of float): Numbers representing the frequencies
                in hz that should be synthesized
        """
        freqs = np.array(freqs_in_hz)
        scaling = 1 / len(freqs)
        sr = int(self.samplerate)
        cps = freqs / sr
        ts = (duration / Seconds(1)) * sr
        ranges = np.array([np.arange(0, ts * c, c) for c in cps])
        raw = (np.sin(ranges * (2 * np.pi)) * scaling).sum(axis=0)
        return AudioSamples(raw, self.samplerate)


class TickSynthesizer(object):
    """
    Synthesize short, percussive, periodic "ticks"

    Args:
        samplerate (SampleRate): the samplerate at which the ticks should be
            synthesized

    Examples:
        >>> import zounds
        >>> synth = zounds.TickSynthesizer(zounds.SR22050())
        >>> samples = synth.synthesize(\
            duration=zounds.Seconds(3), tick_frequency=zounds.Milliseconds(100))
        >>> samples
        AudioSamples([ -3.91624993e-01,  -8.96939666e-01,   4.18165378e-01, ...,
                -4.08054347e-04,  -2.32257899e-04,   0.00000000e+00])

    See Also:
        :class:`SineSynthesizer`
        :class:`NoiseSynthesizer`
        :class:`SilenceSynthesizer`
    """

    def __init__(self, samplerate):
        super(TickSynthesizer, self).__init__()
        self.samplerate = samplerate

    def synthesize(self, duration, tick_frequency):
        """
        Synthesize periodic "ticks", generated from white noise and an envelope

        Args:
            duration (numpy.timedelta64): The total duration of the sound to be
                synthesized
            tick_frequency (numpy.timedelta64): The frequency of the ticking
                sound
        """
        sr = self.samplerate.samples_per_second
        # create a short, tick sound
        tick = np.random.uniform(low=-1., high=1., size=int(sr * .1))
        tick *= np.linspace(1, 0, len(tick))
        # create silence
        samples = np.zeros(int(sr * (duration / Seconds(1))))
        ticks_per_second = Seconds(1) / tick_frequency
        # introduce periodic ticking sound
        step = int(sr // ticks_per_second)
        for i in xrange(0, len(samples), step):
            size = len(samples[i:i + len(tick)])
            samples[i:i + len(tick)] += tick[:size]
        return AudioSamples(samples, self.samplerate)


class NoiseSynthesizer(object):
    """
    Synthesize white noise

    Args:
        samplerate (SampleRate): the samplerate at which the ticks should be
            synthesized

    Examples:
        >>> import zounds
        >>> synth = zounds.NoiseSynthesizer(zounds.SR44100())
        >>> samples = synth.synthesize(zounds.Seconds(2))
        >>> samples
        AudioSamples([ 0.1137964 , -0.02613194,  0.30963904, ..., -0.71398137,
               -0.99840281,  0.74310827])

    See Also:
        :class:`SineSynthesizer`
        :class:`TickSynthesizer`
        :class:`SilenceSynthesizer`
    """

    def __init__(self, samplerate):
        super(NoiseSynthesizer, self).__init__()
        self.samplerate = samplerate

    def synthesize(self, duration):
        """
        Synthesize white noise

        Args:
            duration (numpy.timedelta64): The duration of the synthesized sound
        """
        sr = self.samplerate.samples_per_second
        seconds = duration / Seconds(1)
        samples = np.random.uniform(low=-1., high=1., size=int(sr * seconds))
        return AudioSamples(samples, self.samplerate)


class SilenceSynthesizer(object):
    """
    Synthesize silence

    Args:
        samplerate (SampleRate): the samplerate at which the ticks should be
            synthesized

    Examples:
        >>> import zounds
        >>> synth = zounds.SilenceSynthesizer(zounds.SR11025())
        >>> samples = synth.synthesize(zounds.Seconds(5))
        >>> samples
        AudioSamples([ 0.,  0.,  0., ...,  0.,  0.,  0.])
    """

    def __init__(self, samplerate):
        super(SilenceSynthesizer, self).__init__()
        self.samplerate = samplerate

    def synthesize(self, duration):
        """
        Synthesize silence

        Args:
            duration (numpy.timedelta64): The duration of the synthesized sound
        """
        return AudioSamples.silence(self.samplerate, duration)



import numpy as np
from scipy.spatial.distance import cdist
from index import SearchResults
from random import choice
from zounds.timeseries import ConstantRateTimeSeries
from zounds.nputil import packed_hamming_distance


class BaseBruteForceSearch(object):
    def __init__(self, gen):
        index = []
        self._ids = []
        for _id, example in gen:
            index.append(example)
            crts = ConstantRateTimeSeries(example)
            for ts, _ in crts.iter_slices():
                self._ids.append((_id, ts))
        self.index = np.concatenate(index)

    def search(self, query, n_results=10):
        raise NotImplementedError()

    def random_search(self, n_results=10):
        query = choice(self.index)
        return self.search(query, n_results)


# class BruteForceSearch(object):
#     def __init__(self, gen):
#         index = []
#         self._ids = []
#         for _id, example in gen:
#             index.append(example)
#             crts = ConstantRateTimeSeries(example)
#             for ts, _ in crts.iter_slices():
#                 self._ids.append((_id, ts))
#         self.index = np.concatenate(index)
#
#     def random_search(self, n_results=10):
#         query = choice(self.index)
#         distances = cdist(query[None, ...], self.index)
#         indices = np.argsort(distances[0])[:n_results]
#         return SearchResults(query, (self._ids[i] for i in indices))

class BruteForceSearch(BaseBruteForceSearch):
    def __init__(self, gen):
        super(BruteForceSearch, self).__init__(gen)

    def search(self, query, n_results=10):
        distances = cdist(query[None, ...], self.index)
        indices = np.argsort(distances[0])[:n_results]
        return SearchResults(query, (self._ids[i] for i in indices))


class HammingDistanceBruteForceSearch(BaseBruteForceSearch):
    def __init__(self, gen):
        super(HammingDistanceBruteForceSearch, self).__init__(gen)

    def search(self, query, n_results=10):
        scores = packed_hamming_distance(query, self.index)
        indices = np.argsort(scores)[:n_results]
        return SearchResults(query, (self._ids[i] for i in indices))



from contentrange import RangeUnitUnsupportedException
from api import ZoundsApp
from search import ZoundsSearch
from training_monitor import \
    TrainingMonitorApp, SupervisedTrainingMonitorApp, GanTrainingMonitorApp, \
    TripletEmbeddingMonitorApp
from cli import ObjectStorageSettings, AppSettings, NeuralNetworkTrainingSettings



from torch import nn
from torch.nn import functional as F


class GatedLayer(nn.Module):
    def __init__(
            self,
            layer_type,
            in_channels,
            out_channels,
            kernel,
            stride=1,
            padding=0,
            dilation=1,
            attention_func=F.sigmoid,
            norm=lambda x: x):

        super(GatedLayer, self).__init__()
        self.norm = norm
        self.conv = layer_type(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=False)
        self.gate = layer_type(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=False)
        self.attention_func = attention_func

    def forward(self, x):
        c = self.conv(x)
        c = self.norm(c)
        g = self.gate(x)
        g = self.norm(g)
        out = F.tanh(c) * self.attention_func(g)
        return out


class GatedConvLayer(GatedLayer):
    def __init__(
            self,
            in_channels,
            out_channels,
            kernel,
            stride=1,
            padding=0,
            dilation=1,
            attention_func=F.sigmoid,
            norm=lambda x: x):
        super(GatedConvLayer, self).__init__(
            nn.Conv1d,
            in_channels,
            out_channels,
            kernel,
            stride,
            padding,
            dilation,
            attention_func,
            norm)


class GatedConvTransposeLayer(GatedLayer):
    def __init__(
            self,
            in_channels,
            out_channels,
            kernel,
            stride=1,
            padding=0,
            dilation=1,
            attention_func=F.sigmoid,
            norm=lambda x: x):
        super(GatedConvTransposeLayer, self).__init__(
            nn.ConvTranspose1d,
            in_channels,
            out_channels,
            kernel,
            stride,
            padding,
            dilation,
            attention_func,
            norm)



import numpy as np
from featureflow import Node, NotEnoughData

from zounds.core import ArrayWithUnits
from zounds.nputil import sliding_window
from zounds.timeseries import TimeSlice


def oggvorbis(s):
    """
    This is taken from the ogg vorbis spec
    (http://xiph.org/vorbis/doc/Vorbis_I_spec.html)

    :param s: the total length of the window, in samples
    """
    try:
        s = np.arange(s)
    except TypeError:
        s = np.arange(s[0])

    i = np.sin((s + .5) / len(s) * np.pi) ** 2
    f = np.sin(.5 * np.pi * i)
    return f * (1. / f.max())


class WindowingFunc(object):
    """
    `WindowingFunc` is mostly a convenient wrapper around `numpy's handy
    windowing functions
    <https://docs.scipy.org/doc/numpy/reference/routines.window.html>`_, or any
    function that takes a size parameter and returns a numpy array-like object.

    A `WindowingFunc` instance can be multiplied with a nother array of any size.

    Args:
        windowing_func (function): A function that takes a size parameter, and
            returns a numpy array-like object

    Examples:
        >>> from zounds import WindowingFunc
        >>> import numpy as np
        >>> wf = WindowingFunc(lambda size: np.hanning(size))
        >>> np.ones(5) *  wf
        array([ 0. ,  0.5,  1. ,  0.5,  0. ])
        >>> np.ones(10) * wf
        array([ 0.        ,  0.11697778,  0.41317591,  0.75      ,  0.96984631,
                0.96984631,  0.75      ,  0.41317591,  0.11697778,  0.        ])

    See Also:
        :class:`~IdentityWindowingFunc`
        :class:`~zounds.spectral.OggVorbisWindowingFunc`
        :class:`~zounds.spectral.HanningWindowingFunc`
    """

    def __init__(self, windowing_func=None):
        super(WindowingFunc, self).__init__()
        self.windowing_func = windowing_func
        self._cache = dict()

    def _wdata(self, size):
        if self.windowing_func is None:
            return None
        try:
            return self._cache[size]
        except KeyError:
            window = self.windowing_func(size)
            self._cache[size] = window
            return window

    def __numpy_ufunc__(self, *args, **kwargs):
        # KLUDGE: This seems really odd, but the mere presence of this
        # numpy-specific magic method seems to serve as a hint to to call
        # this instances __rmul__ implementation, instead of doing element-wise
        # multiplication, as per:
        # http://docs.scipy.org/doc/numpy/reference/arrays.classes.html#numpy.class.__numpy_ufunc__
        raise NotImplementedError()

    def __mul__(self, other):
        size = other.shape[-1]
        wdata = self._wdata(size)
        if wdata is None:
            return other
        return wdata * other

    def __rmul__(self, other):
        return self.__mul__(other)


class IdentityWindowingFunc(WindowingFunc):
    """
    An identity windowing function
    """

    def __init__(self):
        super(IdentityWindowingFunc, self).__init__()


class OggVorbisWindowingFunc(WindowingFunc):
    """
    The windowing function described in the `ogg vorbis specification
    <https://xiph.org/vorbis/doc/Vorbis_I_spec.html#x1-230001.3.2>`_
    """

    def __init__(self):
        super(OggVorbisWindowingFunc, self).__init__(windowing_func=oggvorbis)


class HanningWindowingFunc(WindowingFunc):
    """
    A hanning window function
    """

    def __init__(self):
        super(HanningWindowingFunc, self).__init__(windowing_func=np.hanning)


class SlidingWindow(Node):
    """
    `SlidingWindow` is a processing node that provides a very common precursor
    to many frequency domain transforms: a lapped and windowed view of the time-
    domain signal.

    Args:
        wscheme (SampleRate): a sample rate that describes the frequency and
            duration af the sliding window
        wfunc (WindowingFunc): a windowing function to apply to each frame
        needs (Node): A processing node on which this node relies for its data.
            This will generally be a time-domain signal

    Here's how you'd typically see :class:`SlidingWindow` used in a processing
    graph

    .. code:: python

        import zounds

        Resampled = zounds.resampled(resample_to=zounds.SR11025())

        @zounds.simple_in_memory_settings
        class Sound(Resampled):
            windowed = zounds.ArrayWithUnitsFeature(
                zounds.SlidingWindow,
                needs=Resampled.resampled,
                wscheme=zounds.SampleRate(
                    frequency=zounds.Milliseconds(250),
                    duration=zounds.Milliseconds(500)),
                wfunc=zounds.OggVorbisWindowingFunc(),
                store=True)


        synth = zounds.SineSynthesizer(zounds.SR44100())
        samples = synth.synthesize(zounds.Seconds(5), [220., 440., 880.])

        # process the audio, and fetch features from our in-memory store
        _id = Sound.process(meta=samples.encode())
        sound = Sound(_id)

        print sound.windowed.dimensions[0]
        # TimeDimension(f=0.250068024879, d=0.500045346811)
        print sound.windowed.dimensions[1]
        # TimeDimension(f=9.0702947e-05, d=9.0702947e-05)

    See Also:
        :class:`~zounds.spectral.WindowingFunc`
        :class:`~zounds.timeseries.SampleRate`
    """

    def __init__(self, wscheme, wfunc=None, padwith=0, needs=None):
        super(SlidingWindow, self).__init__(needs=needs)
        self._scheme = wscheme
        self._func = wfunc
        self._padwith = padwith
        self._cache = None

    def _first_chunk(self, data):
        if self._padwith:
            padding = np.zeros(
                (self._padwith,) + data.shape[1:], dtype=data.dtype)
            padding_ts = ArrayWithUnits(padding, data.dimensions)
            return padding_ts.concatenate(data)
        else:
            return data

    def _enqueue(self, data, pusher):
        if self._cache is None:
            self._cache = data
        else:
            self._cache = self._cache.concatenate(data)

    def _dequeue(self):

        duration = TimeSlice(duration=self._scheme.duration)
        frequency = TimeSlice(duration=self._scheme.frequency)
        leftover, arr = self._cache.sliding_window_with_leftovers(
            duration,
            frequency,
            dopad=self._finalized)

        if not arr.size:
            raise NotEnoughData()

        self._cache = leftover

        # BUG: Order matters here (try arr * self._func instead)
        # why does that statement result in __rmul__ being called for each
        # scalar value in arr?
        out = (self._func * arr) if self._func else arr
        return out



from util import encode_timedelta, decode_timedelta


class TimeSliceEncoder(object):
    def __init__(self):
        super(TimeSliceEncoder, self).__init__()

    def dict(self, ts):
        return dict(
            start=encode_timedelta(ts.start),
            duration=encode_timedelta(ts.duration))


class TimeSliceDecoder(object):
    def __init__(self):
        super(TimeSliceDecoder, self).__init__()

    def kwargs(self, d):
        return dict(
            start=decode_timedelta(d['start']),
            duration=decode_timedelta(d['duration']))




import unittest2
from persistence import simple_in_memory_settings


@simple_in_memory_settings
class X(object):
    pass


class PersitenceDecoratorTests(unittest2.TestCase):
    def test_in_memory_settings_maintains_class_name(self):
        self.assertEqual('X', X.__name__)
        self.assertNotEqual('zounds.util.persistence', X.__module__)



import numpy as np


def hyperplanes(means, stds, n_planes):
    if len(means) != len(stds):
        raise ValueError('means and stds must have the same length')

    n_features = len(means)
    a = np.random.normal(means, stds, (n_planes, n_features))
    b = np.random.normal(means, stds, (n_planes, n_features))
    plane_vectors = a - b
    return plane_vectors


def simhash(plane_vectors, data):
    output = np.zeros((len(data), len(plane_vectors)), dtype=np.uint8)
    flattened = data.reshape((len(data), -1))
    x = np.dot(plane_vectors, flattened.T).T
    output[np.where(x > 0)] = 1
    return output


def example_wise_unit_norm(x, return_norms=False):
    original_shape = x.shape

    # flatten all dimensions of x, treating the first axis as examples and all
    # other axes as features
    x = x.reshape((len(x), -1))
    norms = np.linalg.norm(x, axis=-1, keepdims=True)
    normed = np.divide(x, norms, where=norms != 0)
    normed = normed.reshape(original_shape)

    if return_norms:
        return normed, norms
    else:
        return normed



import base64
import re
import numpy as np

TIMEDELTA_DTYPE_RE = re.compile(r'\[(?P<dtype>[^\]]+)\]')


def encode_timedelta(td):
    dtype = TIMEDELTA_DTYPE_RE.search(str(td.dtype)).groupdict()['dtype']
    return base64.b64encode(td.astype(np.uint64).tostring()), dtype


def decode_timedelta(t):
    try:
        v = np.frombuffer(base64.b64decode(t[0]), dtype=np.uint64)[0]
        s = t[1]
        return np.timedelta64(long(v), s)
    except IndexError:
        return t



"""
The timeseries module introduces classes for dealing with time as it relates
to audio signals
"""

from duration import \
    Hours, Minutes, Seconds, Milliseconds, Microseconds, Picoseconds, \
    Nanoseconds

from audiosamples import AudioSamples

from samplerate import \
    SR11025, SR16000, SR22050, SR44100, SR48000, SR96000, HalfLapped, \
    audio_sample_rate,Stride, SampleRate, nearest_audio_sample_rate

from timeseries import TimeSlice, TimeDimension

from variablerate import \
    VariableRateTimeSeries, VariableRateTimeSeriesFeature, \
    VariableRateTimeSeriesEncoder, VariableRateTimeSeriesDecoder

from constantrate import ConstantRateTimeSeries

from functional import categorical, inverse_categorical


import torch
from torch import nn
from torch.autograd import Variable
import torch.nn.functional as F


class RawSampleEmbedding(nn.Module):
    """
    Embed raw audio samples after quantizing them and applying a
    softmax/categorical distribution
    """

    def __init__(self, n_categories, embedding_dim, use_cuda=False):
        super(RawSampleEmbedding, self).__init__()
        self.use_cuda = use_cuda
        self.n_categories = n_categories
        self.embedding_dim = embedding_dim
        self.linear = nn.Linear(
            self.n_categories, self.embedding_dim)

    def _variable(self, x, *args, **kwargs):
        v = Variable(x, *args, **kwargs)
        if self.use_cuda:
            v = v.cuda()
        return v

    def _mu_law(self, x):
        m = self._variable(torch.FloatTensor(1))
        m[:] = self.n_categories + 1
        s = torch.sign(x)
        x = torch.abs(x)
        x = s * (torch.log(1 + (self.n_categories * x)) / torch.log(m))
        return x

    def _shift_and_scale(self, x):
        x = x + 1
        x = x * ((self.n_categories + 1) / 2.)
        return x

    def _one_hot(self, x):
        y = self._variable(torch.arange(0, self.n_categories + 1))
        x = -(((x[..., None] - y) ** 2) * 1e12)
        x = F.softmax(x, dim=-1)
        return x

    def categorical(self, x):
        x = x.view(-1)
        x = self._mu_law(x)
        x = self._shift_and_scale(x)
        x = self._one_hot(x)
        return x

    def forward(self, x):
        sample_size = x.shape[-1]

        # one-hot encode the continuous samples
        x = self.categorical(x)

        # embed the categorical variables into a
        # dense vector
        x = self.linear(x)

        # place all embeddings on the unit sphere
        norms = torch.norm(x, dim=-1)
        x = x / norms.view(-1, 1)
        x = x.view(-1, self.embedding_dim, sample_size)
        return x



import unittest
import numpy as np
from npx import windowed, sliding_window, downsample, Growable


class GrowableTest(unittest.TestCase):
    def test_negative_growth_rate(self):
        self.assertRaises(ValueError, lambda: Growable(np.zeros(10), 0, -1))

    def test_initialized_with_zero_elements(self):
        g = Growable(np.zeros(0), 0, 1)
        g.append(1)
        self.assertEqual(1, g.physical_size)
        self.assertEqual(1, g.logical_size)
        self.assertEqual(1, g._data[0])

    def test_logical_data(self):
        g = Growable(np.zeros(10), 0, 1)
        self.assertEqual(0, g.logical_data.shape[0])
        g.extend([1, 2])
        self.assertEqual(2, g.logical_data.shape[0])
        self.assertTrue(np.all(np.array([1, 2]) == g.logical_data))

    def test_append_no_growth(self):
        g = Growable(np.zeros(10), 0, 1)
        g.append(1)
        self.assertEqual(1, g._data[0])
        self.assertEqual(10, g.physical_size)
        self.assertEqual(1, g.logical_size)

    def test_extend_no_growth(self):
        g = Growable(np.zeros(10), 0, 1)
        g.extend(np.ones(3))
        self.assertTrue(np.all(1 == g._data[:3]))
        self.assertEqual(10, g.physical_size)
        self.assertEqual(3, g.logical_size)

    def test_append_growth(self):
        g = Growable(np.zeros(10), 10, 1)
        g.append(1)
        self.assertEqual(20, g.physical_size)
        self.assertEqual(11, g.logical_size)
        self.assertEqual(1, g._data[10])

    def test_extend_growth(self):
        g = Growable(np.zeros(10), 10, 1)
        g.extend(np.ones(5))
        self.assertEqual(20, g.physical_size)
        self.assertEqual(15, g.logical_size)
        self.assertTrue(np.all(1 == g._data[10:15]))

    def test_lt_one_growth_rate_append(self):
        g = Growable(np.zeros(1), 0, .00001)
        g.append(1)
        g.append(2)
        self.assertEqual(2, g.physical_size)
        self.assertEqual(2, g.logical_size)
        self.assertEqual(1, g._data[0])
        self.assertEqual(2, g._data[1])

    def test_multiple_grow_calls_required(self):
        g = Growable(np.zeros(10), 10, .1)
        g.extend(np.ones(4))
        self.assertEqual(14, g.physical_size)
        self.assertEqual(14, g.logical_size)
        self.assertTrue(np.all(1 == g._data[10:14]))

    def test_multidimensional(self):
        g = Growable(np.zeros((10, 3)), 10, 1)
        g.append(1)
        self.assertEqual(20, g.physical_size)
        self.assertEqual(11, g.logical_size)
        self.assertEqual((20, 3), g._data.shape)


class DownsampleTest(unittest.TestCase):
    def test_downsample_single_example_1D(self):
        a = np.ones(10)
        ds = downsample(a, 2)
        self.assertEqual((5,), ds.shape)

    def test_downsample_single_example_2D(self):
        a = np.ones((11, 11))
        ds = downsample(a, 2)
        self.assertEqual((5, 5), ds.shape)

    def test_downsample_multi_example_1D(self):
        a = np.ones((31, 10))
        ds = downsample(a, (2,))
        self.assertEqual((31, 5), ds.shape)

    def test_downsample_multi_example_2D(self):
        a = np.ones((31, 11, 11))
        ds = downsample(a, (2, 2))
        self.assertEqual((31, 5, 5), ds.shape)


class SlidingWindowTest(unittest.TestCase):
    def test_mismatched_dims_ws(self):
        a = np.zeros(10)
        self.assertRaises(ValueError, lambda: sliding_window(a, (1, 2)))

    def test_mismatched_dims_ss(self):
        a = np.zeros(10)
        self.assertRaises(ValueError, lambda: sliding_window(a, 3, (1, 2)))

    def test_windowsize_too_large_1D(self):
        a = np.zeros(10)
        self.assertRaises(ValueError, lambda: sliding_window(a, 11))

    def test_windowsize_too_large_2D(self):
        a = np.zeros((10, 10))
        self.assertRaises(ValueError, lambda: sliding_window(a, (3, 11)))

    def test_1D_no_step_specified(self):
        a = np.arange(10)
        b = sliding_window(a, 3)
        self.assertEqual((3, 3), b.shape)
        self.assertTrue(np.all(b.ravel() == a[:9]))

    def test_1D_with_step(self):
        a = np.arange(10)
        b = sliding_window(a, 3, 1)
        self.assertEqual((8, 3), b.shape)

    def test_1D_flat_nonflat_equivalent(self):
        a = np.zeros(10)
        bflat = sliding_window(a, 3)
        bnonflat = sliding_window(a, 3, flatten=False)
        self.assertEqual(bflat.shape, bnonflat.shape)
        self.assertTrue(np.all(bflat == bnonflat))

    def test_2D_no_step_specified(self):
        a = np.arange(64).reshape((8, 8))
        b = sliding_window(a, (4, 4))
        self.assertEqual((4, 4, 4), b.shape)

    def test_2D_with_step(self):
        a = np.zeros((8, 8))
        b = sliding_window(a, (4, 4), (1, 1))
        self.assertEqual((25, 4, 4), b.shape)

    def test_2D_nonflat_no_step_specified(self):
        a = np.arange(64).reshape((8, 8))
        b = sliding_window(a, (4, 4), flatten=False)
        self.assertEqual((2, 2, 4, 4), b.shape)

    def test_2D_nonflat_with_step(self):
        a = np.zeros((8, 8))
        b = sliding_window(a, (4, 4), (1, 1), flatten=False)
        self.assertEqual((5, 5, 4, 4), b.shape)


class WindowedTest(unittest.TestCase):
    def test_windowsize_ltone(self):
        a = np.arange(10)
        self.assertRaises(ValueError, lambda: windowed(a, 0, 1))

    def test_stepsize_ltone(self):
        a = np.arange(10)
        self.assertRaises(ValueError, lambda: windowed(a, 1, 0))

    def test_no_windowing(self):
        a = np.arange(10)
        l, w = windowed(a, 1, 1)
        self.assertTrue(a is w)
        self.assertEqual(0, l.shape[0])

    def test_drop_samples(self):
        a = np.arange(10)
        l, w = windowed(a, 1, 2)
        self.assertEqual(5, w.shape[0])
        self.assertEqual(0, l.shape[0])

    def test_windowsize_two_stepsize_one_cut(self):
        a = np.arange(10)
        l, w = windowed(a, 2, 1)
        self.assertEqual(1, l.shape[0])
        self.assertEqual((9, 2), w.shape)

    def test_windowsize_two_stepsize_one_pad(self):
        a = np.arange(10)
        l, w = windowed(a, 2, 1, True)
        self.assertEqual(0, l.shape[0])
        self.assertEqual((10, 2), w.shape)

    def test_windowsize_two_stepsize_two_cut(self):
        a = np.arange(10)
        l, w = windowed(a, 2, 2)
        self.assertEqual(0, l.shape[0])
        self.assertEqual((5, 2), w.shape)

    def test_windowsize_two_stepsize_two_pad(self):
        a = np.arange(10)
        l, w = windowed(a, 2, 2, True)
        self.assertEqual(0, l.shape[0])
        self.assertEqual((5, 2), w.shape)

    def test_windowsize_three_stepsize_two_cut(self):
        a = np.arange(10)
        l, w = windowed(a, 3, 2)
        self.assertEqual(2, l.shape[0])
        self.assertEqual((4, 3), w.shape)

    def test_windowsize_three_stepsize_two_pad(self):
        a = np.arange(10)
        l, w = windowed(a, 3, 2, dopad=True)
        self.assertEqual(0, l.shape[0])
        self.assertEqual((5, 3), w.shape)
        self.assertTrue(np.all([0, 0] == w[-1, -1]))

    def test_windowsize_three_stepsize_three_cut(self):
        a = np.arange(10)
        l, w = windowed(a, 3, 3)
        self.assertEqual(1, l.shape[0])
        self.assertEqual((3, 3), w.shape)

    def test_windowsize_three_stepsize_three_pad(self):
        a = np.arange(10)
        l, w = windowed(a, 3, 3, dopad=True)
        self.assertEqual(0, l.shape[0])
        self.assertEqual((4, 3), w.shape)

    def test_windowsize_gt_length_cut(self):
        a = np.arange(5)
        l, w = windowed(a, 6, 1)
        self.assertEqual(5, l.shape[0])
        self.assertEqual(0, w.shape[0])

    def test_windowsize_gt_length_pad(self):
        a = np.arange(5)
        l, w = windowed(a, 6, 1, dopad=True)
        self.assertEqual(0, l.shape[0])
        self.assertEqual((1, 6), w.shape)

    def test_twod_cut(self):
        a = np.arange(20).reshape((10, 2))
        l, w = windowed(a, 3, 2)
        self.assertEqual(2, l.shape[0])
        self.assertEqual((4, 3, 2), w.shape)

    def test_twod_pad(self):
        a = np.arange(20).reshape((10, 2))
        l, w = windowed(a, 3, 2, dopad=True)
        self.assertEqual(0, l.shape[0])
        self.assertEqual((5, 3, 2), w.shape)

    def test_no_stepsize_specified(self):
        a = np.arange(10)
        l, w = windowed(a, 2)
        self.assertEqual(0, l.shape[0])
        self.assertEqual((5, 2), w.shape)

    def test_can_apply_windowed_twice(self):
        a = np.arange(100)
        _, w = windowed(a, 7, 3)
        _, w2 = windowed(w, 3, 1)
        self.assertEqual((3, 7), w2.shape[1:])

    def test_can_apply_windowed_twice_2(self):
        samples = np.random.random_sample(44100)
        _, w = windowed(samples, 512, 256)
        f = np.fft.fft(w)[:, 1:]
        _, w2 = windowed(f, 3, 1)
        self.assertEqual((3, 511), w2.shape[1:])

    def test_can_apply_windowed_to_integer_dtype(self):
        samples = np.zeros(44100).astype(np.int64)
        l, w = windowed(samples, 8192, 4096)
        self.assertEqual(w.dtype, np.int64)
        self.assertEqual(8192, w.shape[1])



import unittest2
from frequencydimension import \
    FrequencyDimensionEncoder, FrequencyDimensionDecoder, \
    LinearScaleEncoderDecoder, GeometricScaleEncoderDecoder, \
    ExplicitScaleEncoderDecoder, ExplicitFrequencyDimensionEncoder, \
    ExplicitFrequencyDimensionDecoder
from zounds.spectral import \
    FrequencyDimension, FrequencyBand, LinearScale, \
    GeometricScale, ExplicitScale, FrequencyScale, ExplicitFrequencyDimension


class ScaleEncodingTests(unittest2.TestCase):
    def test_can_round_trip_linear_scale(self):
        scale = LinearScale(FrequencyBand(20, 4000), n_bands=100)
        encoder_decoder = LinearScaleEncoderDecoder()
        self.assertTrue(encoder_decoder.can_encode(scale))
        encoded = encoder_decoder.encode(scale)
        self.assertTrue(encoder_decoder.can_decode(encoded))
        decoded = encoder_decoder.decode(encoded)
        self.assertEqual(scale, decoded)

    def test_can_round_trip_geometric_scale(self):
        scale = GeometricScale(20, 5000, 0.05, n_bands=100)
        encoder_decoder = GeometricScaleEncoderDecoder()
        self.assertTrue(encoder_decoder.can_encode(scale))
        encoded = encoder_decoder.encode(scale)
        self.assertTrue(encoder_decoder.can_decode(encoded))
        decoded = encoder_decoder.decode(encoded)
        self.assertEqual(scale, decoded)

    def test_can_round_trip_explicit_scale(self):
        scale = ExplicitScale(GeometricScale(20, 5000, 0.05, n_bands=100))
        encoder_decoder = ExplicitScaleEncoderDecoder()
        self.assertTrue(encoder_decoder.can_encode(scale))
        encoded = encoder_decoder.encode(scale)
        self.assertTrue(encoder_decoder.can_decode(encoded))
        decoded = encoder_decoder.decode(encoded)
        self.assertEqual(scale, decoded)


class FrequencyDimensionTests(unittest2.TestCase):
    def setUp(self):
        self.encoder = FrequencyDimensionEncoder()
        self.decoder = FrequencyDimensionDecoder()

    def test_can_round_trip(self):
        band = FrequencyBand(20, 20000)
        scale = LinearScale(band, 50)
        dim = FrequencyDimension(scale)
        encoded = self.encoder.encode(dim)
        decoded = self.decoder.decode(encoded)
        self.assertIsInstance(decoded, FrequencyDimension)
        self.assertEqual(scale, decoded.scale)

    def test_can_round_trip_specific_scale_type(self):
        band = FrequencyBand(20, 20000)
        scale = LinearScale(band, 50)
        dim = FrequencyDimension(scale)
        encoded = self.encoder.encode(dim)
        decoded = self.decoder.decode(encoded)
        self.assertIsInstance(decoded.scale, LinearScale)
        self.assertEqual(scale, decoded.scale)

    def test_can_round_trip_geometric_scale(self):
        scale = GeometricScale(20, 5000, bandwidth_ratio=0.01, n_bands=100)
        dim = FrequencyDimension(scale)
        encoded = self.encoder.encode(dim)
        decoded = self.decoder.decode(encoded)
        self.assertIsInstance(decoded.scale, GeometricScale)
        self.assertEqual(scale, decoded.scale)

    def test_can_round_trip_explicit_scale(self):
        scale = ExplicitScale(
            GeometricScale(20, 5000, bandwidth_ratio=0.01, n_bands=100))
        dim = FrequencyDimension(scale)
        encoded = self.encoder.encode(dim)
        decoded = self.decoder.decode(encoded)
        self.assertIsInstance(decoded.scale, ExplicitScale)
        self.assertEqual(scale, decoded.scale)

    def test_raises_when_encountering_unknown_scale(self):
        band = FrequencyBand(20, 20000)
        scale = FrequencyScale(band, 50)
        dim = FrequencyDimension(scale)
        self.assertRaises(NotImplementedError, lambda: self.encoder.encode(dim))


class ExplicitFrequencyDimensionTests(unittest2.TestCase):

    def setUp(self):
        self.encoder = ExplicitFrequencyDimensionEncoder()
        self.decoder = ExplicitFrequencyDimensionDecoder()

    def test_can_round_trip(self):
        scale = GeometricScale(20, 5000, 0.05, 3)
        slices = [slice(0, 10), slice(10, 100), slice(100, 1000)]
        dim = ExplicitFrequencyDimension(scale, slices)
        encoded = self.encoder.encode(dim)
        decoded = self.decoder.decode(encoded)
        self.assertIsInstance(decoded, ExplicitFrequencyDimension)
        self.assertEqual(dim, decoded)




import unittest2
from identitydimension import IdentityDimensionEncoder, IdentityDimensionDecoder
from zounds.core import IdentityDimension, Dimension


class OtherDimension(Dimension):
    pass


class IdentityDimensionTests(unittest2.TestCase):
    def setUp(self):
        self.encoder = IdentityDimensionEncoder()
        self.decoder = IdentityDimensionDecoder()

    def test_identity_encoder_matches(self):
        self.assertTrue(self.encoder.matches(IdentityDimension()))

    def test_identity_encoder_does_not_match(self):
        self.assertFalse(self.encoder.matches(OtherDimension()))

    def test_identity_encoder_encodes(self):
        encoded = self.encoder.encode(IdentityDimension())
        self.assertIsInstance(encoded, dict)
        self.assertEqual(IdentityDimension.__name__, encoded['type'])
        self.assertEqual({}, encoded['data'])

    def test_identity_decoder_matches(self):
        encoded = {'type': IdentityDimension.__name__, 'data': {}}
        self.assertTrue(self.decoder.matches(encoded))

    def test_identity_decoder_does_not_match(self):
        encoded = {'type': OtherDimension.__name__, 'data': {}}
        self.assertFalse(self.decoder.matches(encoded))

    def test_identity_decoder_decodes(self):
        decoded = self.decoder.decode(
                {'type': IdentityDimension.__name__, 'data': {}})
        self.assertIsInstance(decoded, IdentityDimension)

    def test_roundtrip(self):
        encoded = self.encoder.encode(IdentityDimension())
        decoded = self.decoder.decode(encoded)
        self.assertIsInstance(decoded, IdentityDimension)



import numpy as np
import unittest2
from functional import \
    fft, stft, apply_scale, frequency_decomposition, phase_shift, rainbowgram, \
    fir_filter_bank
from zounds.core import ArrayWithUnits, IdentityDimension
from zounds.synthesize import \
    SilenceSynthesizer, TickSynthesizer, SineSynthesizer, FFTSynthesizer
from zounds.timeseries import SR22050, Seconds, Milliseconds, TimeDimension, \
    TimeSlice, AudioSamples
from zounds.spectral import \
    HanningWindowingFunc, FrequencyDimension, LinearScale, GeometricScale, \
    ExplicitFrequencyDimension, FrequencyBand
from matplotlib import cm


class FIRFilterBankTests(unittest2.TestCase):
    def test_has_correct_dimensions(self):
        samplerate = SR22050()
        scale = GeometricScale(
            start_center_hz=20,
            stop_center_hz=10000,
            bandwidth_ratio=0.2,
            n_bands=100)
        scale.ensure_overlap_ratio(0.5)
        taps = 256
        filter_bank = fir_filter_bank(scale, taps, samplerate, np.hanning(100))
        self.assertEqual((len(scale), taps), filter_bank.shape)
        self.assertEqual(FrequencyDimension(scale), filter_bank.dimensions[0])
        self.assertEqual(TimeDimension(*samplerate), filter_bank.dimensions[1])


class FrequencyDecompositionTests(unittest2.TestCase):
    def test_can_decompose_audio_samples(self):
        samples = AudioSamples.silence(SR22050(), Seconds(1))
        bands = frequency_decomposition(samples, [64, 128, 256, 512, 1024])
        expected_td = TimeDimension(samples.end, samples.end)
        self.assertEqual(expected_td, bands.dimensions[0])
        self.assertIsInstance(bands.dimensions[1], ExplicitFrequencyDimension)

    def test_can_decompose(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(9999))
        wscheme = sr.windowing_scheme(8192, 4096)
        duration = TimeSlice(wscheme.duration)
        frequency = TimeSlice(wscheme.frequency)
        _, windowed = samples.sliding_window_with_leftovers(
            duration, frequency, dopad=True)
        fa = frequency_decomposition(
            windowed, [32, 64, 128, 256, 512, 1024, 2048, 4096])
        self.assertEqual(windowed.dimensions[0], fa.dimensions[0])
        self.assertIsInstance(
            fa.dimensions[1], ExplicitFrequencyDimension)


class FFTTests(unittest2.TestCase):
    def test_can_pad_for_better_frequency_resolution(self):
        samples = SilenceSynthesizer(SR22050()).synthesize(Milliseconds(2500))
        windowsize = TimeSlice(duration=Milliseconds(200))
        stepsize = TimeSlice(duration=Milliseconds(100))
        _, windowed = samples.sliding_window_with_leftovers(
            windowsize=windowsize, stepsize=stepsize, dopad=True)
        coeffs = fft(windowed, padding_samples=1024)
        self.assertIsInstance(coeffs, ArrayWithUnits)
        self.assertEqual(2, len(coeffs.dimensions))
        self.assertEqual(windowed.dimensions[0], coeffs.dimensions[0])
        self.assertIsInstance(coeffs.dimensions[1], FrequencyDimension)
        expected_size = ((windowed.shape[-1] + 1024) // 2) + 1
        self.assertEqual(expected_size, coeffs.shape[-1])

    def test_can_take_fft_of_1d_signal(self):
        samples = SilenceSynthesizer(SR22050()).synthesize(Milliseconds(2500))
        coeffs = fft(samples)
        self.assertIsInstance(coeffs, ArrayWithUnits)
        self.assertEqual(1, len(coeffs.dimensions))
        self.assertIsInstance(coeffs.dimensions[0], FrequencyDimension)

    def test_can_take_fft_of_2d_stacked_signal(self):
        samples = SilenceSynthesizer(SR22050()).synthesize(Milliseconds(2500))
        windowsize = TimeSlice(duration=Milliseconds(200))
        stepsize = TimeSlice(duration=Milliseconds(100))
        _, windowed = samples.sliding_window_with_leftovers(
            windowsize=windowsize, stepsize=stepsize, dopad=True)
        coeffs = fft(windowed)
        self.assertIsInstance(coeffs, ArrayWithUnits)
        self.assertEqual(2, len(coeffs.dimensions))
        self.assertEqual(windowed.dimensions[0], coeffs.dimensions[0])
        self.assertIsInstance(coeffs.dimensions[1], FrequencyDimension)


class PhaseShiftTests(unittest2.TestCase):
    def _mean_squared_error(self, x, y):
        l = min(len(x), len(y))
        return ((x[:l] - y[:l]) ** 2).mean()

    def test_1d_phase_shift_returns_correct_size(self):
        samplerate = SR22050()
        samples = SineSynthesizer(samplerate) \
            .synthesize(Milliseconds(5500), [220, 440, 880])
        coeffs = fft(samples)
        shifted = phase_shift(
            coeffs=coeffs,
            samplerate=samplerate,
            time_shift=Milliseconds(5500),
            frequency_band=FrequencyBand(50, 5000))
        self.assertEqual(coeffs.shape, shifted.shape)

    def test_can_phase_shift_1d_signal(self):
        samplerate = SR22050()
        samples = SineSynthesizer(samplerate) \
            .synthesize(Milliseconds(5000), [220, 440, 880])
        coeffs = fft(samples)
        shifted = phase_shift(coeffs, samplerate, Milliseconds(10))
        new_samples = np.fft.irfft(shifted, norm='ortho')
        self.assertNotEqual(0, self._mean_squared_error(samples, new_samples))

    def test_can_phase_shift_1d_signal_180_degrees(self):
        samplerate = SR22050()
        samples = SineSynthesizer(samplerate) \
            .synthesize(Seconds(1), [110, 220, 440, 880])
        coeffs = fft(samples)
        shifted = phase_shift(
            coeffs=coeffs,
            samplerate=samplerate,
            time_shift=-Milliseconds(1000),
            frequency_band=FrequencyBand(50, 5000))
        new_samples = np.fft.irfft(shifted, norm='ortho')
        self.assertAlmostEqual(
            0, self._mean_squared_error(samples, new_samples), 1)

    def test_2d_phase_shift_returns_correct_shape(self):
        samplerate = SR22050()
        samples = SineSynthesizer(samplerate) \
            .synthesize(Milliseconds(2500), [220, 440, 880])
        windowsize = TimeSlice(duration=Milliseconds(200))
        stepsize = TimeSlice(duration=Milliseconds(100))
        _, windowed = samples.sliding_window_with_leftovers(
            windowsize=windowsize, stepsize=stepsize, dopad=True)
        coeffs = fft(windowed)
        shifted = phase_shift(
            coeffs=coeffs,
            samplerate=samplerate,
            time_shift=Milliseconds(40),
            frequency_band=FrequencyBand(50, 5000))
        self.assertEqual(coeffs.shape, shifted.shape)

    def test_can_phase_shift_2d_signal(self):
        samplerate = SR22050()
        samples = SineSynthesizer(samplerate) \
            .synthesize(Milliseconds(2500), [220, 440, 880])
        windowsize = TimeSlice(duration=Milliseconds(200))
        stepsize = TimeSlice(duration=Milliseconds(100))
        _, windowed = samples.sliding_window_with_leftovers(
            windowsize=windowsize, stepsize=stepsize, dopad=True)
        coeffs = fft(windowed)
        shifted = phase_shift(coeffs, samplerate, Milliseconds(40))
        synth = FFTSynthesizer()
        new_samples = synth.synthesize(shifted).squeeze()
        self.assertNotEqual(0, self._mean_squared_error(samples, new_samples))

    def test_can_phase_shift_2d_signal_180_degrees(self):
        samplerate = SR22050()
        samples = SineSynthesizer(samplerate) \
            .synthesize(Milliseconds(2500), [220, 440, 880])
        windowsize = TimeSlice(duration=Milliseconds(200))
        stepsize = TimeSlice(duration=Milliseconds(100))
        _, windowed = samples.sliding_window_with_leftovers(
            windowsize=windowsize, stepsize=stepsize, dopad=True)
        coeffs = fft(windowed)
        shifted = phase_shift(
            coeffs=coeffs,
            samplerate=samplerate,
            time_shift=Milliseconds(100))
        synth = FFTSynthesizer()
        new_samples = synth.synthesize(shifted).squeeze()
        self.assertAlmostEqual(
            0, self._mean_squared_error(samples, new_samples), 1)

    def test_raises_value_error_when_specified_axis_not_frequency_dim(self):
        samplerate = SR22050()
        samples = SineSynthesizer(samplerate) \
            .synthesize(Milliseconds(2500), [220, 440, 880])
        self.assertRaises(
            ValueError,
            lambda: phase_shift(samples, samplerate, Milliseconds(10)))


class STFTTests(unittest2.TestCase):
    def test_has_correct_number_of_bins(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(6666))
        wscheme = sr.windowing_scheme(512, 256)
        tf = stft(samples, wscheme, HanningWindowingFunc())
        self.assertEqual(tf.shape[1], 257)

    def test_has_correct_dimensions(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(6666))
        wscheme = sr.windowing_scheme(512, 256)
        tf = stft(samples, wscheme, HanningWindowingFunc())
        self.assertIsInstance(tf, ArrayWithUnits)
        self.assertEqual(2, len(tf.dimensions))
        self.assertIsInstance(tf.dimensions[0], TimeDimension)
        self.assertEqual(tf.dimensions[0].samplerate, wscheme)
        self.assertIsInstance(tf.dimensions[1], FrequencyDimension)
        self.assertIsInstance(tf.dimensions[1].scale, LinearScale)


class RainbowgramTests(unittest2.TestCase):
    def test_should_have_correct_shape_and_dimensions(self):
        samplerate = SR22050()
        samples = SineSynthesizer(samplerate).synthesize(Milliseconds(8888))
        wscheme = samplerate.windowing_scheme(256, 128)
        tf = stft(samples, wscheme, HanningWindowingFunc())
        rg = rainbowgram(tf, cm.rainbow)
        self.assertEqual(3, rg.ndim)
        self.assertEqual(tf.shape + (3,), rg.shape)
        self.assertEqual(3, len(rg.dimensions))
        self.assertEqual(tf.dimensions[0], rg.dimensions[0])
        self.assertEqual(tf.dimensions[1], rg.dimensions[1])
        self.assertEqual(rg.dimensions[2], IdentityDimension())
        self.assertEqual(3, rg.shape[-1])


class ApplyScaleTests(unittest2.TestCase):
    def test_apply_scale_to_self_is_identity_function(self):
        samplerate = SR22050()
        samples = SineSynthesizer(samplerate).synthesize(Milliseconds(8888))
        wscheme = samplerate.windowing_scheme(256, 128)
        tf = stft(samples, wscheme, HanningWindowingFunc())
        scale = tf.dimensions[-1].scale
        transformed = apply_scale(tf, scale, HanningWindowingFunc())
        self.assertEqual(tf.shape, transformed.shape)
        self.assertEqual(tf.dimensions[0], transformed.dimensions[0])
        self.assertEqual(tf.dimensions[1], transformed.dimensions[1])

    def test_has_correct_shape(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(9999))
        wscheme = sr.windowing_scheme(256, 128)
        scale = GeometricScale(50, sr.nyquist, 0.4, 32)
        scale.ensure_overlap_ratio()
        tf = stft(samples, wscheme, HanningWindowingFunc())
        geom = apply_scale(tf, scale, window=HanningWindowingFunc())
        self.assertEqual(tf.shape[:-1] + (len(scale),), geom.shape)

    def test_has_correct_dimensions(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(9999))
        wscheme = sr.windowing_scheme(256, 128)
        scale = GeometricScale(50, sr.nyquist, 0.4, 32)
        scale.ensure_overlap_ratio()
        tf = stft(samples, wscheme, HanningWindowingFunc())
        geom = apply_scale(tf, scale, window=HanningWindowingFunc())
        self.assertIsInstance(geom, ArrayWithUnits)
        self.assertEqual(2, len(geom.dimensions))
        self.assertIsInstance(geom.dimensions[0], TimeDimension)
        self.assertEqual(geom.dimensions[0].samplerate, wscheme)
        self.assertIsInstance(geom.dimensions[1], FrequencyDimension)
        self.assertEqual(scale, geom.dimensions[1].scale)

    def test_preserves_time_dimension(self):
        sr = SR22050()
        samples = TickSynthesizer(sr).synthesize(
            Milliseconds(10000), Milliseconds(1000))
        wscheme = sr.windowing_scheme(256, 128)
        scale = GeometricScale(50, sr.nyquist, 0.4, 32)
        scale.ensure_overlap_ratio()
        tf = stft(samples, wscheme, HanningWindowingFunc())
        geom = apply_scale(tf, scale, window=HanningWindowingFunc())

        # get the loudness envelope of each
        tf_envelope = np.abs(tf.real).sum(axis=1)
        geom_envelope = geom.sum(axis=1)

        tf_zeros = np.where(tf_envelope == 0)
        geom_zeros = np.where(geom_envelope == 0)

        np.testing.assert_allclose(tf_zeros, geom_zeros)



class Trainer(object):
    def __init__(self, epochs, batch_size, checkpoint_epochs=1):
        super(Trainer, self).__init__()
        self.checkpoint_epochs = checkpoint_epochs
        self.batch_size = batch_size
        self.epochs = epochs
        self._batch_complete_callbacks = dict()
        self._current_epoch = 0

    def train(self):
        raise NotImplemented()

    def register_batch_complete_callback(self, callback):
        self._batch_complete_callbacks[id(callback)] = callback

    def unregister_batch_complete_callback(self, callback):
        try:
            del self._batch_complete_callbacks[id(callback)]
        except KeyError:
            # the callback was never registered
            pass

    def on_batch_complete(self, *args, **kwargs):
        for callback in self._batch_complete_callbacks.itervalues():
            callback(*args, **kwargs)




from __future__ import division
from duration import Picoseconds
from collections import namedtuple
import numpy as np
from duration import Seconds

Stride = namedtuple('Stride', ['frequency', 'duration'])


class SampleRate(object):
    """
    `SampleRate` describes the constant frequency at which samples are taken
    from a continuous signal, and the duration of each sample.

    Instances of this class could describe an audio sampling rate (e.g. 44.1kHz)
    or the strided windows often used in short-time fourier transforms

    Args:
        frequency (numpy.timedelta64): The frequency at which the signal is
            sampled
        duration (numpy.timedelta64): The duration of each sample

    Raises:
        ValueError: when frequency or duration are less than or equal to zero

    Examples:
        >>> from zounds import Seconds, SampleRate
        >>> sr = SampleRate(Seconds(1), Seconds(2))
        >>> sr.frequency
        numpy.timedelta64(1,'s')
        >>> sr.duration
        numpy.timedelta64(2,'s')
        >>> sr.overlap
        numpy.timedelta64(1,'s')
        >>> sr.overlap_ratio
        0.5

    See Also:
        :class:`SR96000`
        :class:`SR48000`
        :class:`SR44100`
        :class:`SR22050`
        :class:`SR11025`
    """

    def __init__(self, frequency, duration):
        if frequency.astype(np.int) <= 0:
            raise ValueError('frequency must be positive')
        if duration.astype(np.int) <= 0:
            raise ValueError('duration must be positive')

        self.frequency = frequency
        self.duration = duration
        super(SampleRate, self).__init__()

    def __str__(self):
        f = self.frequency / Seconds(1)
        d = self.duration / Seconds(1)
        return '{self.__class__.__name__}(f={f}, d={d})'.format(**locals())

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        return iter((self.frequency, self.duration))

    def __len__(self):
        return 2

    def __eq__(self, other):
        return \
            self.frequency == other.frequency \
            and self.duration == other.duration

    @property
    def overlap(self):
        """
        For sampling schemes that overlap, return a :class:`numpy.timedelta64`
        instance representing the duration of overlap between each sample
        """
        return self.duration - self.frequency

    @property
    def overlap_ratio(self):
        """
        For sampling schemes that overlap, return the ratio of overlap to
        sample duration
        """
        return self.overlap / self.duration

    @property
    def samples_per_second(self):
        return int(Picoseconds(int(1e12)) / self.frequency)

    @property
    def nyquist(self):
        return self.samples_per_second // 2

    def __mul__(self, other):
        try:
            if len(other) == 1:
                other *= 2
        except TypeError:
            other = (other, other)

        freq = self.frequency * other[0]
        duration = (self.frequency * other[1]) + self.overlap
        new = SampleRate(freq, duration)
        print self, new
        return new

    def discrete_samples(self, ts):
        td = ts.dimensions[0]
        windowsize = np.round((self.duration - td.overlap) / td.frequency)
        stepsize = np.round(self.frequency / td.frequency)
        return int(stepsize), int(windowsize)

    def resample(self, ratio):
        orig_freq = Picoseconds(int(self.frequency / Picoseconds(1)))
        orig_duration = Picoseconds(int(self.duration / Picoseconds(1)))
        f = orig_freq * ratio
        d = orig_duration * ratio
        return SampleRate(f, d)


class AudioSampleRate(SampleRate):
    def __init__(self, samples_per_second, suggested_window, suggested_hop):
        self.suggested_hop = suggested_hop
        self.suggested_window = suggested_window
        self.one_sample = Picoseconds(int(1e12)) // samples_per_second
        super(AudioSampleRate, self).__init__(self.one_sample, self.one_sample)

    def __int__(self):
        return self.samples_per_second

    def half_lapped(self):
        return SampleRate(
            self.one_sample * self.suggested_hop,
            self.one_sample * self.suggested_window)

    def windowing_scheme(self, duration_samples, frequency_samples=None):
        frequency_samples = frequency_samples or duration_samples
        return SampleRate(
            self.frequency * frequency_samples,
            self.duration * duration_samples)


class SR96000(AudioSampleRate):
    """
    A :class:`SampleRate` representing the common audio sampling rate 96kHz

    Examples:
        >>> from zounds import SR96000
        >>> sr = SR96000()
        >>> sr.samples_per_second
        96000
        >>> int(sr)
        96000
        >>> sr.nyquist
        48000
    """

    def __init__(self):
        super(SR96000, self).__init__(96000, 4096, 2048)


class SR48000(AudioSampleRate):
    """
    A :class:`SampleRate` representing the common audio sampling rate 48kHz

    Examples:
        >>> from zounds import SR48000
        >>> sr = SR48000()
        >>> sr.samples_per_second
        48000
        >>> int(sr)
        48000
        >>> sr.nyquist
        24000
    """

    def __init__(self):
        super(SR48000, self).__init__(48000, 2048, 1024)


class SR44100(AudioSampleRate):
    """
    A :class:`SampleRate` representing the common audio sampling rate 44.1kHz

    Examples:
        >>> from zounds import SR44100
        >>> sr = SR44100()
        >>> sr.samples_per_second
        44100
        >>> int(sr)
        44100
        >>> sr.nyquist
        22050
    """

    def __init__(self):
        super(SR44100, self).__init__(44100, 2048, 1024)


class SR22050(AudioSampleRate):
    """
    A :class:`SampleRate` representing the common audio sampling rate 22.025kHz

    Examples:
        >>> from zounds import SR22050
        >>> sr = SR22050()
        >>> sr.samples_per_second
        22050
        >>> int(sr)
        22050
        >>> sr.nyquist
        11025
    """

    def __init__(self):
        super(SR22050, self).__init__(22050, 1024, 512)


class SR16000(AudioSampleRate):
    """
        A :class:`SampleRate` representing the common audio sampling rate 16kHz

        Examples:
            >>> from zounds import SR16000
            >>> sr = SR16000()
            >>> sr.samples_per_second
            16000
            >>> int(sr)
            16000
            >>> sr.nyquist
            8000
        """

    def __init__(self):
        super(SR16000, self).__init__(16000, 512, 256)


class SR11025(AudioSampleRate):
    """
    A :class:`SampleRate` representing the common audio sampling rate 11.025kHz

    Examples:
        >>> from zounds import SR11025
        >>> sr = SR11025()
        >>> sr.samples_per_second
        11025
        >>> int(sr)
        11025
        >>> sr.nyquist
        5512
    """

    def __init__(self):
        super(SR11025, self).__init__(11025, 512, 256)


_samplerates = (
    SR96000(), SR48000(), SR44100(), SR22050(), SR16000(), SR11025())


def audio_sample_rate(samples_per_second):
    for sr in _samplerates:
        if samples_per_second == sr.samples_per_second:
            return sr
    raise ValueError(
        '{samples_per_second} is an invalid sample rate'.format(**locals()))


def nearest_audio_sample_rate(samples_per_second):
    samplerates = np.array([s.samples_per_second for s in _samplerates])
    diffs = np.abs(samples_per_second - samplerates)
    return _samplerates[np.argmin(diffs)]


class HalfLapped(SampleRate):
    def __init__(self, window_at_44100=2048, hop_at_44100=1024):
        one_sample_at_44100 = Picoseconds(int(1e12)) / 44100.
        window = one_sample_at_44100 * window_at_44100
        step = one_sample_at_44100 * hop_at_44100
        super(HalfLapped, self).__init__(step, window)



import requests
import urlparse
import os


def ensure_local_file(remote_url, local_path, chunksize=4096):
    parsed = urlparse.urlparse(remote_url)
    filename = os.path.split(parsed.path)[-1]
    local = os.path.join(local_path, filename)

    if not os.path.exists(local):
        with open(local, 'wb') as f:
            resp = requests.get(remote_url, stream=True)
            total_bytes = int(resp.headers['Content-Length'])
            for i, chunk in enumerate(resp.iter_content(chunk_size=chunksize)):
                f.write(chunk)
                progress = ((i * chunksize) / float(total_bytes)) * 100
                print '{remote_url} {progress:.2f}% complete'.format(**locals())

    return local
